// btest.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2024-12-15 16:30:16 tac>
//
// 度数集計表を入力し、出現率一定の検定または二項検定と出現率の信頼区間を
// 計算する。度数集計表は、計数値解析コマンド用の標準入力形式に準じた形式
// になっている必要がある。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W btest btest.C
//              -I/usr/local/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "btest  Version 7.42          Copyleft(Vo\\oV) 2024 杉本典夫",
  "書式:btest  [<オプション> …]  [<度数集計表ファイル> …]",
  "機能:度数集計表に対して出現率の検定をするんや。",
  "<オプション>",
  "  -cc[=<信頼係数>]:信頼係数を指定するんや。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -ccs[={1|2}]:信頼区間の種類を指定するんや。",
  "    1:片側  2:両側(オプション省略時)",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込むんや。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するんや。(-h:-h=man)",
  "  -p=<分類1の理論確率>[,<分類2の理論確率>,…]:理論確率を指定するんや。",
  "    オプション省略時:-p=1/分類数,1/分類数,…  (0<理論確率<1)",
  "  -res=[-][{def|all|bt[c]|hwt[{f|o}]|m[{b|f|s}]|tab}][,…]",
  "    :解析結果の出力項目を指するんや。",
  "    -:-を付けた項目を出力しないんや(-res=-:全ての項目を出力しないんや)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    bt[{c|o}]:出現率の検定(c:連続修正有 o:連続修正無)",
  "    hwt[{f|o}]:Hardy-Weinberg平衡の検定",
  "      f:Fisherの正確検定も出力 o:Fisherの正確検定を出力しない",
  "    m[{b|f|s}]:多重比較(b:Bonferroni型 f:Fisher型 s:Scheffe型)",
  "    tab:度数集計表",
  "    オプション省略時:データに合わせて自動判定するんや"
};

const string Title = "=== 分類データの出現率比較 ==="; // 表題

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  testtype vbt;                 // 出現率の検定の種類
  double vcc;                   // 信頼係数
  int vccs;                     // 信頼区間の種類
  testtype vhwt;                // Hardy-Weinberg平衡の検定の種類
  mctype vm;                    // 2分類ごとの多重比較法
  lstr plst;                    // 理論確率リスト
  vecdbl vpvec;                 // 理論確率ベクトル
  tabtype vtab;                 // 度数集計表出力フラッグ
  lstr vreslst;                 // 結果出力リスト
  mapsb vresmap;                // 結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  testtype bt() const           // 出現率の検定の種類を返す
  { return vbt; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  int ccs() const               // 信頼区間の種類を返す
  { return vccs; }
  testtype hwt() const         // Hardy-Weinberg平衡の検定の種類を返す
  { return vhwt; }
  mctype m() const              // 2分類ごとの多重比較法を返す
  { return vm; }
  double p0(int pno) const      // 理論確率を返す
  { return (pno < (int)(vpvec.size())) ? vpvec[pno] : BLANKf; }
  const vecdbl& pvec() const    // 理論確率ベクトルを返す
  { return vpvec; }
  tabtype tab() const           // 度数集計表の種類を返す
  { return vtab; }
  bool res(const string& code); // 解析結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int nr);            // オプションをチェックする
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vbt(testcor), vcc(CCdef), vccs(2), vhwt(testetc), vm(mcbon), plst(), vpvec(1), vtab(tabfreq), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
  vpvec[0] = 1.0;
}

bool option::res(const string& code) // 結果出力フラッグを返す
{
  mapsb_citr respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "ccs")     // 信頼区間の指定
  {
    if (opt[1].empty() || opt[1] == "2") vccs = 2;
    else if (opt[1] == "1") vccs = 1;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "p")       // 理論確率指定
  {
    if (strsplit(plst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(int nr)       // オプションをチェックする
{
  if (nr < 1) return 0;         // 理論確率自動設定
  if (nr <= (int)(plst.size())) return disperr(vpname, "理論確率が多すぎるんや！");
  vpvec.resize(nr + 1);
  vpvec[0] = 1.0;               // 0:不等確率,1:等確率
  double p = 0.0, pt = 0.0;
  int i = 1;
  lstr_citr pitr = plst.begin();
  while (pitr != plst.end() && i < nr)
  {
    p = strtodf(*pitr);
    pt += p;
    if (p <= 0.0 || 1.0 <= p || pt > 1.0)
    { return disperr(vpname, "理論確率が間違っとるんや！"); }
    else
    {
      if (i > 1 && p != vpvec[i - 1]) vpvec[0] = 0.0;
      vpvec[i] = p;
      ++pitr;  ++i;
    }
  }
  p = (1.0 - pt) / (double)(nr + 1 - i); // 残りの理論確率
  if (i > 1 && p != vpvec[i - 1]) vpvec[0] = 0.0;
  while (i <= nr)
  {
    vpvec[i] = p;
    ++i;
  }
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラー無し
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["bt"] = vresmap["tab"] = true;
  vresmap["hwt"] = vresmap["m"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["bt"] = vresmap["tab"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr respos = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(resstr.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def") // 省略時出力
      { vresmap["def"] = vresmap["bt"] = vresmap["tab"] = true; }
      else if (resstr == "all") // 全て出力
      {
        vresmap["def"] = vresmap["bt"] = vresmap["hwt"] = vresmap["m"] = vresmap["tab"] = true;
        vhwt = testetc;         // Fisherの正確検定も出力する
      }
      else if (resstr == "btc") // 連続修正有り
      { vresmap["bt"] = true;  vbt = testcor; }
      else if (resstr == "bto") // 連続修正無し
      { vresmap["bt"] = true;  vbt = testorg; }
      else if (resstr == "hwtf") // Fisherの正確検定も出力する
      { vresmap["hwt"] = true;  vhwt = testetc; }
      else if (resstr == "hwto") // Fisherの正確検定は出力しない
      { vresmap["hwt"] = true;  vhwt = testorg; }
      else if (resstr[0] == 'm' && resstr.size() > 1)
      {
        switch (resstr[1])
        {
        case 'b' : vm = mcbon;
          break;
        case 'f' : vm = mcfis;
          break;
        case 's' : vm = mcsch;
          break;
        default : return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！");
        }
        vresmap["m"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

class btestclass                // btestメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_table(ostream& fout, option& opt, const ftable& ftab) const;
  // 度数集計表を出力する
  void put_btest(ostream& fout, option& opt, const ftable& ftab) const;
  // 二項検定の結果を出力する
  void put_ctest(ostream& fout, option& opt, const ftable& ftab) const;
  // χ2乗検定の結果を出力する
  void put_cmcon(ostream& fout, option& opt, mctype mc, const ftable& ftab) const;
  // 出現率の多重比較結果を出力する
  void put_hwtest(ostream& fout, option& opt, const ftable& ftab) const;
  // Hardy-Weinberg平衡の検定の結果を出力する
  ldbl hwtest(double& p1, double& p2, double& vr, double& csqr, const ftable& ftab) const;
  // 遺伝子に関するHardy-Weinberg平衡の検定(Fisherの両側直接確率を返す)
public:
  btestclass() : fname()        // コンストラクタ
  {}
  ~btestclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // 度数集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst) const;
  // 結果を出力する
};

void btestclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put(fout);
}

void btestclass::put_table(ostream& fout, option& opt, const ftable& ftab) const
// 度数集計表を出力する
{
  string codei(ftab.ccode(0));
  int npad = codei.size() - strclen(codei, CHARwide);
  fout << "\n" << left << setw(ftab.cwide(0) + npad) << codei << right;
  codei = ftab.ccode(1);
  npad = codei.size() - strclen(codei, CHARwide);
  fout << setw(ftab.cwide(1) + npad) << codei;
  fout << " (出現率";
  int w = ftab.hwide() + ftab.cwide(1) + 10;
  if (opt.ccs() != 0)
  {
    fout << setw(5) << opt.cc() * 100.0;
    if (opt.ccs() == 1) fout << "%片側信頼区間 ";
    else fout << "%両側信頼区間 ";
    w += 19;
  }
  fout << "%)\n";
  putline(fout, w);
  int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
  fout.precision(3);  fout.setf(ios::fixed);
  int nt = ftab.ftotal();
  double cc = opt.cc();
  if (opt.ccs() == 1) cc = (opt.cc() > 0.5) ? opt.cc() * 2.0 - 1.0 : 0.0;
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    codei = ftab.rcode(i);
    npad = codei.size() - strclen(codei, CHARwide);
    fout << left << setw(ftab.cwide(0) + npad) << codei << right;
    int ni = ftab.f(i, 1);
    double pr = (double)ni / (double)nt;
    fout << setw(ftab.cwide(1)) << ni;
    fout << " (" << setw(7) << pr * 100.0;
    if (opt.ccs() != 0)
    {
      double plow = 0.0, pup = 1.0;
      calbexci(plow, pup, ni, nt, cc);
      if (opt.ccs() == 1)
      {
        if (pr < 0.5) plow = 0.0;
        else pup = 1.0;
      }
      fout << setw(8) << plow * 100.0 << " --" << setw(8) << pup * 100.0;
    }
    fout << ")\n";
  }
  putline(fout, w);
  codei = "合計";
  npad = codei.size() - strclen(codei, CHARwide);
  fout << left << setw(ftab.hwide() + npad) << codei << right;
  fout << setw(ftab.cwide(1)) << ftab.fctotal(1);
  fout << " (" << setw(7) << 100.0;
  fout << ")\n";
  fout.precision(prec0);  fout.unsetf(ios::fixed);
}

void btestclass::put_btest(ostream& fout, option& opt, const ftable& ftab) const
// 二項検定の結果を出力する
{
  ldbl pr1 = binom(ftab.f(1, 1), ftab.f(2, 1), opt.p0(1));
  ldbl pr = (pr1 < 0.5L) ? 2.0L * pr1 : 1.0L; // 両側p値
  fout << "\n・二項検定  有意確率:片側p=" << pr1 << pind(pr1) << " 両側p=" << pr << pind(pr);
}

void btestclass::put_ctest(ostream& fout, option& opt, const ftable& ftab) const
// χ2乗検定の結果を出力する
{
  int df = ftab.nr() - 1;
  double csqr = ftest(opt.pvec(), ftab, (opt.bt() == testcor) ? true : false);
  ldbl pr = ctop(csqr, df);
  fout << "\n・χ2乗検定(";
  if (opt.bt() == testcor) fout << "修正有,";
  fout << ftab.nr() << "×1)  χ2乗=" << csqr << "  自由度=" << df << "  有意確率:p=" << pr << pind(pr) << "\n";
  fout << "  理論確率:";
  for (size_t i = 1; i <= ftab.nr(); ++i) fout << ftab.rcode(i) << "=" << opt.p0(i) << "  ";
  fout << endl;
}

void btestclass::put_cmcon(ostream& fout, option& opt, mctype mc, const ftable& ftab) const
// 出現率の多重比較結果を出力する
{
  double f0 = 0.5 * (double)(ftab.nr()) / double(ftab.ftotal());
  int df = 1, dc = ftab.nr() * (ftab.nr() - 1) / 2;
  string method("Bonferroni");
  switch (mc)
  {
  case mcfis : method = "Fisher";  dc = 1;  df = 1;
    break;
  case mcsch : method = "Scheffe";
    df = ftab.nr() - 1;  dc = 1;
    break;
  }
  fout << "\n・" << method << "型多重比較(" << method << " type multiple comparison)\n";
  fout << "コード  :  出現率 - コード  :  出現率      χ2乗値  自由度   有意確率p値\n";
  constexpr int LENline = 75;   // 区切り線の長さ
  putline(fout, LENline);
  for (size_t i = 1; i < ftab.nr(); ++i)
  {
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    int fi = ftab.f(i, 1);
    double pri = (double)fi / (double)(ftab.ftotal()) * 100.0;
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      double prj = (double)(ftab.f(j, 1)) / (double)(ftab.ftotal()) * 100.0;
      double csqr = f0 * (double)(fi - ftab.f(j, 1)) * (fi - ftab.f(j, 1));
      double pr = calmcpr(mc, csqr, ftab.nr(), df, 0, dc);
      string codej(ftab.rcode(j));
      int npad = codej.size() - strclen(codej, CHARwide);
      fout << left << setw(8 + npad) << codej;
      fout << ":" << right << setw(8) << prj;
      fout << left << " - ";
      fout << setw(8 + npadi) << codei;
      fout << ":" << right << setw(8) << pri;
      fout << setw(12) << csqr << setw(8) << df;
      fout << setw(14) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, LENline);
}

void btestclass::put_hwtest(ostream& fout, option& opt, const ftable& ftab) const
// Hardy-Weinberg平衡の検定の結果を出力する
{
  double p1 = 0.0, p2 = 0.0, vr = 0.0, csqr = 0.0;
  ldbl prfis = hwtest(p1, p2, vr, csqr, ftab);
  double nn = 2 * ftab.ftotal();
  double r2 = csqr / nn;
  double r = sqrt(r2);
  fout << "\n・Hardy-Weinberg平衡の検定\n";
  fout << "  推定出現確率:遺伝子1=" << p1 << "  遺伝子2=" << p2 << "  分散=" << vr << "\n";
  fout << "  理論出現確率:" << ftab.rcode(1) << "=" << p1 * p1;
  fout << "  " << ftab.rcode(2) << "=" << 2.0 * p1 * p2;
  fout << "  " << ftab.rcode(3) << "=" << p2 * p2 << "\n";
  fout << "  Cramerの連関係数:V=" << r << "  寄与率:V2乗=" << r2 << "\n";
  double csqrl = 0.0, csqru = 0.0;
  calcsqrci(csqrl, csqru, opt.cc(), 1, csqr);
  double rl = sgn(r) * sqrt(csqrl / nn);
  double ru = sgn(r) * sqrt(csqru / nn);
  if (rl < -1.0) rl = -1.0;
  if (rl > 1.0) rl = 0.0;
  if (ru < -1.0) ru = -1.0;
  if (ru > 1.0) ru = 1.0;
  ldbl pr = ctop(csqr, 1);
  fout << "  連関係数の" << opt.cc() * 100.0 << "%信頼区間:下限=" << rl << "  上限=" << ru << "\n";
  fout << "  χ2乗=" << csqr << "  自由度=1  有意確率:p=" << pr << pind(pr) << "\n";
  if (opt.hwt() == testetc) fout << "  Fisherの正確検定:両側有意確率p=" << prfis << pind(prfis) << "\n";
}

ldbl btestclass::hwtest(double& p1, double& p2, double& vr, double& csqr, const ftable& ftab) const
// 遺伝子に関するHardy-Weinberg平衡の検定(Fisherの両側直接確率を返す)
{
  p1 = (double)(2 * ftab.f(1, 1) + ftab.f(2, 1)) / (double)(2 * ftab.ftotal());
  p2 = 1.0 - p1;
  vr = p1 * p2 / (double)(2 * ftab.ftotal());
  vecdbl pvec(4);
  pvec[0] = 0.0;
  pvec[1] = p1 * p1;  pvec[2] = 2.0 * p1 * p2;  pvec[3] = p2 * p2;
  ftable ftab2(ftab);           // 計算用
  ftab2.f(1, 1 ,2 * ftab.f(1, 1));  ftab2.f(2, 1, 2 * ftab.f(2, 1));
  ftab2.f(3, 1, 2 * ftab.f(3, 1));
  csqr = ftest(pvec, ftab2, false);
  ldbl pr = fisher(2 * ftab.f(1, 1), ftab.f(2, 1), ftab.f(2, 1), 2 * ftab.f(3, 1));
  return (pr < 0.5L) ? 2.0L * pr : 1.0L; // 両側確率を返す
}

int btestclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("実行時エラー！");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void btestclass::put(option& opt, listidd& idlst, listftab& ftablst) const
// 結果を出力する
{
  listidd_itr iitr = idlst.begin();
  for (auto& ftabi : ftablst)
  {
    if (opt.check(ftabi.nr()) != 0) throw option_error("実行時オプションエラー");
    if (opt.res("res"))
    {
      put_id(cout, *iitr);
      if (ftabi.nr() > 0)
      {
        if (opt.res("tab")) put_table(cout, opt, ftabi);
        if (opt.res("bt"))      // 出現率の検定出力
        {
          if (ftabi.nr() == 2) put_btest(cout, opt, ftabi);
          put_ctest(cout, opt, ftabi);
        }
        if (ftabi.nr() > 2 && opt.res("m") && opt.p0(0) == 1.0) put_cmcon(cout, opt, opt.m(), ftabi);
        if (ftabi.nr() == 3 && opt.res("hwt")) put_hwtest(cout, opt, ftabi);
      }
      cout << "\n" << endl;
    }
    ++iitr;
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  btestclass btest;             // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          btest.get(nullptr, opt, idlst, ftablst);
          btest.put(opt, idlst, ftablst);
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        btest.get(argv[i], opt, idlst, ftablst);  fok = false;
        btest.put(opt, idlst, ftablst);
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
