// cctest.C Version 7.41  Copyleft (c) 杉本典夫
// Time-stamp: <2023-12-15 10:37:33 tac>
//
// クロス集計表を入力し、累積χ2乗検定と最大χ2乗法による検定を行う。
// 単純な2×2χ2乗検定(多重比較にしない間違った手法!)も行うことができる。
// クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W cctest cctest.C -I/usr/local/include
//               -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "cctest  Version 7.41         Copyleft(Vo\\oV) 2023 杉本典夫",
  "書式:cctest  [<オプション> …]  [<クロス集計表ファイル> …]",
  "機能:クロス集計表に対して累積χ2乗検定を行います。",
  "<オプション>",
  "  -cc[=<信頼係数>]:出現率の差の信頼係数を指定します。(0≦信頼係数<1)",
  "  -ct[={-|+}]:累積χ2乗検定の出力指定です。(-ct:-ct=+)",
  "  -cx[={-|b|f}]:2×2分割表の多重比較法を指定します。(-cx:-cx=b)",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込みます。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示します。(-h:-h=man)",
  "  -m[={-|b|f|s}]:群間の多重比較法を指定します。(-m:-m=b)",
  "    -:出力せず(オプション省略時)  b:Bonferroni型  f:Fisher型  s:Scheffe型",
  "  -tab[={-|+|p|pc|pr|pt}]:クロス集計表の出力指定です。(-tab:-tab=p)",
  "    -:出力せず  +:出力  p:行合計に対する%も出力(オプション省略時)",
  "    pc:列合計に対する%も出力  pr:pと同じ  pt:全合計に対する%も出力"
};

const string Title = "=== 順序分類データの群間比較 ==="; // 表題

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  bool vct;                     // 累積χ2乗検定出力フラッグ
  mctype vcx;                   // 2×2分割表の多重比較法
  mctype vm;                    // 群間の多重比較法
  tabtype vtab;                 // 度数集計表の種類
  bool vym;                     // 多重比較のYatesの修正有無フラッグ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  bool ct() const               // 累積χ2乗検定出力フラッグを返す
  { return vct; }
  mctype cx() const             // 2×2分割表の多重比較法を返す
  { return vcx; }
  mctype m() const              // 群間の多重比較法を返す
  { return vm; }
  tabtype tab() const           // 度数集計表出力フラッグを返す
  { return vtab; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vct(true), vcx(mcnon), vm(mcnon), vtab(tabfprr)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "ct")      // 累積χ2乗検定の出力指定
  {
    if (opt[1].empty() || opt[1] == "+") vct = true;
    else if (opt[1] == "-") vct = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "cx")      // 2×2分割表の多重比較法指定
  {
    if (opt[1].empty() || opt[1] == "b") vcx = mcbon;
    else if (opt[1] == "-") vcx = mcnon;
    else if (opt[1] == "f") vcx = mcfis;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "m")       // 群間の多重比較法指定
  {
    if (opt[1].empty() || opt[1] == "b") vm = mcbon;
    else if (opt[1] == "-") vm = mcnon;
    else if (opt[1] == "f") vm = mcfis;
    else if (opt[1] == "s") vm = mcsch;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "tab")     // 度数表の出力指定
  {
    if (opt[1].empty() || opt[1] == "p" || opt[1] == "pr") vtab = tabfprr;
    else if (opt[1] == "-") vtab = tabnon;
    else if (opt[1] == "+") vtab = tabfreq;
    else if (opt[1] == "pc") vtab = tabfprc;
    else if (opt[1] == "pt") vtab = tabfprt;
    else err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

class cctestclass               // cctestメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_cctest(ostream& fout, double r, double csqr, double v) const;
  // 累積χ2乗検定の結果を出力する
  void put_ccmcon(ostream& fout, const option& opt, const ftable& ftab, double r) const;
  // 累積χ2乗法の多重比較
  void put_cmax(ostream& fout, const option& opt, const ftable& ftab) const;
  // 累積度数表に対する最大χ2乗検定(Bonferroni型)または単純χ2乗検定(Fisher型)
public:
  cctestclass() : fname()       // コンストラクタ
  {}
  ~cctestclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // 度数集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst) const;
  // 結果を出力する
};

void cctestclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "群項目(縦)      :" << id.item(i) << "\n";
      break;
    case avalcol : fout << "順序分類項目(横):" << id.item(i) << "\n";
      break;
    default : id.put_item(fout, i);
    }
  }
}

void cctestclass::put_cctest(ostream& fout, double r, double csqr, double v) const
// 累積χ2乗検定の結果を出力する
{
  ldbl pr = cctop(csqr, v);
  fout << "\n・累積χ2乗検定    Σχ^2と自由度の補正係数:ρ=" << r << "\n";
  fout << "  ρ・Σχ^2=" << csqr << "  ρ・自由度=" << v << "  有意確率:p=" << pr << pind(pr) << endl;
}

void cctestclass::put_ccmcon(ostream& fout, const option& opt, const ftable& ftab, double r) const
// 累積χ2乗法の多重比較
{
  double nt = (double)(ftab.ftotal());
  double v = r * (double)(ftab.nc() - 1);
  int dc = ftab.nr() * (ftab.nr() - 1) / 2;
  string method("Bonferroni");
  switch (opt.m())
  {
  case mcfis : method = "Fisher";  dc = 1;
    break;
  case mcsch : method = "Scheffe";
    v = r * (double)(ftab.nr() - 1) * (double)(ftab.nc() - 1);  dc = 1;
  }
  fout << "\n・群の" << method << "型多重比較(" << method << " type multiple comparison)\n";
  fout << "群         - 群            ρ・Σχ2乗値   ρ・自由度   有意確率p値\n";
  constexpr int LENline = 67;   // 区切り線の長さ
  putline(fout, LENline);
  for (size_t i = 1; i < ftab.nr(); ++i)
  {
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    double xi = (double)(ftab.frtotal(i));
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      double xj = (double)(ftab.frtotal(j)), csqr = 0.0;
      for (size_t k = 1; k < ftab.nc(); ++k)
      {
        double dp = (double)(ftab.f(i, k)) / xi - (double)(ftab.f(j, k)) / xj;
        csqr += dp * dp * nt / (double)(ftab.fctotal(k)) / (nt - (double)(ftab.fctotal(k)));
      }
      csqr *= r * nt * xi * xj / (xi + xj);
      ldbl pr = cctop(csqr, v) * dc;
      if (pr > 1.0L) pr = 1.0L;
      string codej(ftab.rcode(j));
      int npad = codej.size() - strclen(codej, CHARwide);
      fout << left << setw(11 + npad) << codej;
      fout << "- ";
      fout << setw(10 + npadi) << codei;
      fout << right << setw(14) << csqr << setw(12) << v;
      fout << setw(14) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, LENline);
}

void cctestclass::put_cmax(ostream& fout, const option& opt, const ftable& ftab) const
// 累積度数表に対する最大χ2乗検定(Bonferroni型)または単純χ2乗検定(Fisher型)
{
  int dc = ftab.nc() - 1;
  string method("Bonferroni型多重比較(最大χ2乗検定)");
  if (opt.cx() == mcfis) { method = "Fisher型多重比較(単純χ2乗検定)";  dc = 1; }
  fout << "\n○累積度数分布表に対する" << method << "\n";
  if (opt.tab() != tabnon) putftable(fout, ftab, opt.tab());
  fout << "\n";
  double w = (double)(ftab.ftotal());
  int df = ftab.nr() - 1;       // χ2乗検定自由度
  bool putci = (df == 1) ? true : false;
  int llen = 42;                // 区切り線の長さ
  fout << right << "コード  ";
  if (putci)
  {
    fout << "  出現率の差" << setw(5) << opt.cc() * 100.0 << "%CI";
    fout << "下限        上限";
    llen += 36;
  }
  fout << "     χ2乗値 自由度 有意確率p値\n";
  putline(fout, llen);
  for (size_t j = 1; j < ftab.nc(); ++j)
  {
    double csqr = 0.0;
    for (size_t i = 1; i <= ftab.nr(); ++i) csqr += (double)(ftab.f(i, j)) / (double)(ftab.frtotal(i)) * (double)(ftab.f(i, j));
    csqr = (csqr - (double)(ftab.fctotal(j)) / w * (double)(ftab.fctotal(j))) * w * w / (double)(ftab.fctotal(j)) / (w - (double)(ftab.fctotal(j)));
    ldbl pr = ctop(csqr, df) * dc;
    if (pr > 1.0L) pr = 1.0L;
    string code(ftab.ccode(j));
    strtrm(code);
    int npad = code.size() - strclen(code, CHARwide);
    fout << left << setw(8 + npad) << code;
    if (putci)
    {
      double d = 0.0;
      double ci = caldpci(d, opt.cx(), opt.cc(), ftab.nc(), df, dc, ftab.f(1, j), ftab.frtotal(1) - ftab.f(1, j), ftab.f(2, j), ftab.frtotal(2) - ftab.f(2, j), false);
      fout << right << setw(12) << d << setw(12) << d - ci << setw(12) << d + ci;
    }
    fout << right << setw(11) << csqr << setw(7) << df;
    fout << setw(12) << pr << pind(pr) << "\n";
  }
  putline(fout, llen);
}

int cctestclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("実行時エラー！");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void cctestclass::put(option& opt, listidd& idlst, listftab& ftablst) const
// 結果を出力する
{
  listidd::iterator iitr = idlst.begin();
  for (listftab::iterator fitr = ftablst.begin(); fitr != ftablst.end(); ++iitr, ++fitr)
  {
    put_id(cout, *iitr);
    fitr->trm();                // 行・列の合計が0のものを削除する
    if (fitr->nr() > 0)
    {
      if (opt.tab() != tabnon) putftable(cout, *fitr, opt.tab());
      if (fitr->nr() > 1 && fitr->nc() > 1)
      {
        fitr->fcum();           // 累積度数表にする
        double csqr = 0.0, v = 0.0;
        double r = cctest(csqr, v, *fitr);
        if (opt.ct()) put_cctest(cout, r, csqr, v);
        if (fitr->nr() > 2 && opt.m() != mcnon) put_ccmcon(cout, opt, *fitr, r);
        if (opt.cx() != mcnon) put_cmax(cout, opt, *fitr);
      }
    }
    cout << "\n" << endl;
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  cctestclass cctest;           // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          cctest.get(nullptr, opt, idlst, ftablst);
          cctest.put(opt, idlst, ftablst);
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        cctest.get(argv[i], opt, idlst, ftablst);  fok = false;
        cctest.put(opt, idlst, ftablst);
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
