// ctest.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2024-12-15 16:38:54 tac>
//
// クロス集計表を入力し、各種の統計量を求めて推定と検定を行う。
// クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W ctest ctest.C
//              -I/usr/local/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "ctest  Version 7.42          Copyleft(Vo\\oV) 2024 杉本典夫",
  "書式:ctest  [<オプション> …]  [<クロス集計表ファイル> …]",
  "機能:クロス集計表に関する各種指標の検定をするんじゃ。",
  "<オプション>",
  "  -al[=<有意水準>]:検出力分析の有意水準αの指定じゃ。(0<α<1,-al:-al=0.05)",
  "  -cc[=<信頼係数>]:信頼係数の指定じゃ。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -dm[=<検出差>]:検出力分析の検出差の指定じゃ。(-dm:-dm=<標本出現率の差>)",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込むんじゃ。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するんじゃ。(-h:-h=man)",
  "  -res=[-][{def|all|b2[{1|n}]|bc|lt|m[{a|d}][{b|d|f|s|t}]",
  "    |or[{o|w}]|pa|r|rd|rr|st|tab[{fr|pc|pr|pt}]|ym}][,…]",
  "    :解析結果の出力項目を指定します。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    b2[{1|n}]:mx2単独分割表(1:最初のコードとの分割表のみ",
  "      n:全ての2種類のコードの分割表)  bc:mx2累積分割表",
  "    lt:尤度比検定",
  "    m[{a|d}][{b|d|f|s|t}]:多重比較(ma:全ての2群比較 md:群1と他の群の比較)",
  "    or[{o|w}]:オッズ比の検定(o:Woolfの修正無 w:Woolfの修正有)",
  "    pa:検出力分析  r:Cramerの連関係数  rd:リスク差の検定  rr:リスク比の検定",
  "    st:交差度数の優劣比較(McNemerの検定)",
  "    tab[={fr|pc|pr|pt}]:クロス集計表(fr:度数だけ pc:列合計に対する％も出力",
  "      pr:行合計に対する％も出力 pt:全合計に対する％も出力)",
  "    ym:多重比較におけるYatesの修正",
  "    オプション省略時:データに合わせて自動判定"
};

const string Title = "=== 分類データの群間比較 ==="; // 表題

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double val;                   // 有意水準
  int vb2;                      // m×2単独分割表の出力フラッグ
  double vcc;                   // 信頼係数
  double vdm;                   // 検出差
  mctype vma;                   // 全ての2群ごとの多重比較法
  mctype vmd;                   // 対照群と他の群の多重比較法
  testtype vodr;                // オッズ比出力フラッグ
  double vpo;                   // 検出力
  tabtype vtab;                 // 度数集計表の種類
  lstr vreslst;                 // 結果出力リスト
  mapsb vresmap;                // 結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double al() const             // 有意水準を返す
  { return val; }
  int b2() const                // m×2単独分割表の出力フラッグを返す
  { return vb2; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  double dm() const             // 検出差を返す
  { return vdm; }
  mctype ma() const             // 全ての2群ごとの多重比較法を返す
  { return vma; }
  mctype md() const             // 対照群と他の群の多重比較法を返す
  { return vmd; }
  testtype odr() const          // オッズ比出力フラッグを返す
  { return vodr; }
  double po() const             // 検出力を返す
  { return vpo; }
  tabtype tab() const           // 度数集計表出力フラッグを返す
  { return vtab; }
  bool res(const string& code); // 解析結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int nftab);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), val(ALdef), vb2(2), vcc(CCdef), vdm(0.0), vma(mctuk), vmd(mcdun), vodr(testaut), vpo(POdef), vtab(tabfprr), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 結果出力フラッグを返す
{
  mapsb_citr respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "al")           // 有意水準の指定
  {
    val = opt[1].empty() ? ALdef : strtodf(opt[1]);
    if (val <= 0.0 || 1.0 <= val) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "cc")      // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "dm")      // 検出差の指定
  {
    vdm = opt[1].empty() ? 0.0 : strtodf(opt[1]);
    if (vdm < 0.0 || 1.0 <= vdm) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "po")      // 検出力の指定
  {
    vpo = opt[1].empty() ? POdef : strtodf(opt[1]);
    if (vpo <= 0.0 || 1.0 <= vpo) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(int nftab)    // オプションのセットとチェック
{
  if (nftab < 1) return disperr(vpname, "クロス集計表がない！");
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["rd"] = vresmap["tab"] = true;
  vresmap["b2"] = vresmap["bc"] = vresmap["lt"] = vresmap["ma"] = vresmap["md"] = vresmap["or"] = vresmap["pa"] = vresmap["r"] = vresmap["rr"] = vresmap["st"] = vresmap["ym"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["rd"] = vresmap["tab"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr respos = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(resstr.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def") // 省略時出力
      { vresmap["def"] = vresmap["rd"] = vresmap["tab"] = true; }
      else if (resstr == "all") // 全て出力
      {
        vresmap["def"] = vresmap["b2"] = vresmap["bc"] = vresmap["lt"] = vresmap["ma"] = vresmap["md"] = vresmap["or"] = vresmap["pa"] = vresmap["r"] = vresmap["rd"] = vresmap["rr"] = vresmap["st"] = vresmap["tab"] = vresmap["ym"] = true;
        vb2 = 2;  vma = mctuk;  vmd = mcdun;  vodr = testaut;  vtab = tabfprr;
      }
      else if (resstr == "b21") // 最初のコードとのmx2単独分割表のみ
      { vresmap["b2"] = true;  vb2 = 1; }
      else if (resstr == "b2n") // mx2単独分割表
      { vresmap["b2"] = true;  vb2 = 2; }
      else if (resstr == "m")      // -res=mと指定された時
      { vresmap["ma"] = vresmap["md"] = true; }
      else if (resstr[0] == 'm' && resstr[1] == 'a' && resstr.size() > 2)
      {
        switch (resstr[2])
        {
        case 'b' : vma = mcbon;
          break;
        case 'f' : vma = mcfis;
          break;
        case 's' : vma = mcsch;
          break;
        case 't' : vma = mctuk;
          break;
        default : return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！");
        }
        vresmap["ma"] = true;
      }
      else if (resstr[0] == 'm' && resstr[1] == 'd' && resstr.size() > 2)
      {
        switch (resstr[2])
        {
        case 'b' : vmd = mcbon;
          break;
        case 'd' : vmd = mcdun;
          break;
        case 'f' : vmd = mcfis;
          break;
        case 's' : vmd = mcsch;
          break;
        default : return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！");
        }
        vresmap["md"] = true;
      }
      else if (resstr == "oro") // Woolfの修正無し
      { vresmap["or"] = true;  vodr = testorg; }
      else if (resstr == "orw") // Woolfの修正有り
      { vresmap["or"] = true;  vodr = testaut; }
      else if (resstr == "tabfr") // 度数のみ
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (resstr == "tabpc") // 列合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (resstr == "tabpr") // 行合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (resstr == "tabpt") // 全合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (respos == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

class ctestclass                // ctestメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_cramer(ostream& fout, option& opt, const ftable& ftab) const;
  // Cramerの連関係数を出力する
  void put_ltest(ostream& fout, const ftable& ftab) const;
  // 尤度比検定結果を出力する
  void put_rd(ostream& fout, option& opt, const ftable& ftab) const;
  // リスク差を出力する
  void put_pa(ostream& fout, option& opt, const ftable& ftab) const;
  // 検出力分析結果を出力する
  void put_rr(ostream& fout, option& opt, const ftable& ftab) const;
  // リスク比を出力する
  bool cal_rr(double& rr, double& b, double& v, double& csqr, option& opt, const ftable& ftab) const;
  // リスク比または陽性尤度比を出力する
  void put_odr(ostream& fout, option& opt, const ftable& ftab) const;
  // オッズ比を出力する
  testtype cal_odr(double& odr, double& b, double& v, double& csqr, option& opt, const ftable& ftab) const;
  // オッズ比とその検定を行う
  void put_st(ostream& fout, option& opt, const ftable& ftab) const;
  // 交差度数の優劣比較の結果を出力する
  void put_stpci(ostream& fout, option& opt, int f1, int f2) const;
  // 交差度数の出現率の信頼区間出力
  void put_mcon(ostream& fout, option& opt, const ftable& ftab) const;
  // 群間の多重比較結果を出力する
  void put_macon(ostream& fout, mctype mc, option& opt, const ftable& ftab) const;
  // 全ての2群間の多重比較結果を出力する
  void put_mdcon(ostream& fout, mctype mc, option& opt, const ftable& ftab) const;
  // 対照群と他の群の多重比較結果を出力する
  void put_bun(ostream& fout, option& opt, idd& id, const ftable& ftab) const;
  // m×2累積分割表を出力する
  void put_bun2(ostream& fout, option& opt, idd& id, const ftable& ftab) const;
  // m×2単独分割表を出力する
  void put_cmcon(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const;
  // 出現率の差の多重比較結果を出力する
  void put_cmconci(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const;
  // 出現率の差の同時信頼区間を出力する
  void put_mpa(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const;
  // 出現率の差の検出力分析結果を出力する
  void put_rrmcon(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const;
  // リスク比の多重比較結果を出力する
  void put_rrmconci(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const;
  // リスク比の同時信頼区間を出力する
  void put_odrmcon(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const;
  // オッズ比の多重比較結果を出力する
  void put_odrmconci(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const;
  // オッズ比の同時信頼区間を出力する
public:
  ctestclass() : fname()        // コンストラクタ
  {}
  ~ctestclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // 度数集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst) const;
  // 結果を出力する
};

void ctestclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "群項目(縦)  :" << id.item(i) << "\n";
      break;
    case avalcol : fout << "分類項目(横):" << id.item(i) << "\n";
      break;
    default : id.put_item(fout, i);
    }
  }
}

void ctestclass::put_cramer(ostream& fout, option& opt, const ftable& ftab) const
// Cramerの連関係数を出力する
{
  double csqr = 0.0;
  int df = 0;
  testtype ctype = ctest(csqr, df, ftab, false); // 修正無しのχ2乗検定
  ldbl pr = ctop(csqr, df);
  double s = (double)(min(ftab.nr(), ftab.nc()) - 1);
  double r2 = csqr / (double)(ftab.ftotal()) / s;
  double r = sqrt(r2);
  fout << "\n・Cramerの連関係数:V=" << r << "  寄与率:V2乗=" << r2 << "\n";
  double csqrl = 0.0, csqru = 0.0;
  calcsqrci(csqrl, csqru, opt.cc(), df, csqr);
  double rl = sgn(r) * sqrt(csqrl / (double)(ftab.ftotal()) / s);
  double ru = sgn(r) * sqrt(csqru / (double)(ftab.ftotal()) / s);
  if (rl < -1.0) rl = -1.0;
  if (rl > 1.0) rl = 0.0;
  if (ru < -1.0) ru = -1.0;
  if (ru > 1.0) ru = 1.0;
  fout << "  連関係数の" << opt.cc() * 100.0 << "%信頼区間:下限=" << rl << "  上限=" << ru << "\n";
  fout << "  χ2乗(" << ftab.nr() << "×" << ftab.nc() << ")=" << csqr << "  自由度=" << df << "  有意確率:p=" << pr << pind(pr) << endl;
}

void ctestclass::put_ltest(ostream& fout, const ftable& ftab) const
// 尤度比検定結果を出力する
{
  double lr = 0.0, nt = (double)(ftab.ftotal());
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double xi = (double)(ftab.frtotal(i));
    for (size_t j = 1; j <= ftab.nc(); ++j)
    {
      if (ftab.f(i, j) > 0)
      {
        double xij = (double)(ftab.f(i, j));
        lr += xij * log(xij / xi * nt / (double)(ftab.fctotal(j)));
      }
    }
  }
  double csqr = 2.0 * lr;
  lr /= nt;
  int df = (ftab.nr() - 1) * (ftab.nc() - 1);
  ldbl pr = ctop(csqr, df);
  fout << "\n・尤度比(likelihood ratio):LR=" << exp(lr) << "  ln(LR)=" << lr << "\n";
  fout << "  χ2乗(" << ftab.nr() << "×" << ftab.nc() << ")=" << csqr << "  自由度=" << df << "  有意確率:p=" << pr << pind(pr) << endl;
}

void ctestclass::put_rd(ostream& fout, option& opt, const ftable& ftab) const
// リスク差を出力する
{
  bool cor = true;              // 連続修正は常に有り
  double d = 0.0, pci = 0.0;
  fout << "\n・出現率の差";
  if (opt.tab() == tabfprc)
  {
    fout << "=陽性率差:Dp=";
    pci = caldpci(d, mcfis, opt.cc(), 2, 1, 1, ftab.f(1, 1), ftab.f(2, 1), ftab.f(1, 2), ftab.f(2, 2), cor);
  }
  else
  {
    fout << "=リスク差(risk difference):RD=";
    pci = caldpci(d, mcfis, opt.cc(), 2, 1, 1, ftab.f(1, 1), ftab.f(1, 2), ftab.f(2, 1), ftab.f(2, 2), cor);
  }
  fout << d << "\n";
  fout << "  出現率の差の" << opt.cc() * 100.0 << "%信頼区間=" << d - pci << "-" << d + pci << "\n";
  double csqr = 0.0;
  int df = 0;
  testtype ctype = ctest(csqr, df, ftab, cor);
  ldbl pr = ctop(csqr, df);
  fout << "  χ2乗検定(連続修正有):χ2乗=" << csqr << "  自由度=" << df << "  有意確率:p=" << pr << pind(pr) << "\n";
  ldbl pr1 = fisher(ftab.f(1, 1), ftab.f(1, 2), ftab.f(2, 1), ftab.f(2, 2));
  pr = (pr1 < 0.5L) ? 2.0L * pr1 : 1.0L; // 両側p値
  fout << "  Fisherの正確検定 有意確率:片側p=" << pr1 << pind(pr1) << " 両側p=" << pr << pind(pr) << endl;
}

void ctestclass::put_pa(ostream& fout, option& opt, const ftable& ftab) const
// 検出力分析結果を出力する
{
  int n12 = ftab.f(1, 2), n22 = ftab.f(2, 2);
  int n1 = ftab.frtotal(1), n2 = ftab.frtotal(2);
  if (n1 > n2)                  // 小さい方をn1にする
  {
    n12 = ftab.f(2, 2);  n22 = ftab.f(1, 2);
    n1 = ftab.frtotal(2);  n2 = ftab.frtotal(1);
  }
  double p1 = (double)n12 / (double)n1;
  double p2 = (double)n22 / (double)n2;
  double dm = (opt.dm() > 0.0) ? opt.dm() : fabs(p1 - p2);
  if (sgn(dm) == 0) dm = 0.1; // 出現率の差が0の時は0.1にする
  double s = (double)n2 / (double)n1;
  double pi = (p1 + s * p2) / (s + 1.0);
  double v = pi * (1.0 - pi), za = ptoz(opt.al());
  double tmp1 = za * sqrt((s + 1.0) * v);
  double tmp2 = sqrt(s * p1 * (1.0 - p1) + p2 * (1.0 - p2));
  double n1t = (double)n1 * 0.9;
  double tmp3 = 1.0 + sqrt(1.0 + 2.0 * (s + 1.0) / s / n1t / dm);
  n1t = 4.0 * (double)n1 / tmp3 / tmp3;
  double z2b = (sgn(tmp2) == 0) ? 0.0 : (sqrt(s * dm * dm * n1t) - tmp1) / tmp2;
  ldbl po = sndis(z2b);
  fout << "\n・検出力分析:有意水準=" << opt.al() << "  χ2乗検定(連続修正有)\n";
  fout << "  検出差" << dm << "の時の検出力=" << po << "(β=" << 1.0L - po << ")\n";
  double b = 1.0 - opt.po();
  z2b = (b <= 0.5) ? ptoz(2.0 * b) : -ptoz(2.0 * opt.po());
  double dp = (za + z2b) * sqrt((1.0 + s) / s * v / (double)n1);
  fout << "  検出力=" << opt.po() << "(β=" << b << ")の時の検出差=" << dp << "\n";
  s = 1.0;                      // 2群の例数比を1にする
  pi = (p1 + s * p2) / (s + 1.0);
  v = pi * (1.0 - pi);
  double nsqrt = (za * sqrt((1.0 + s) * v) + z2b * sqrt(s * p1 * (1.0 - p1) + p2 * (1.0 - p2))) / dm;
  double n0 = nsqrt * nsqrt / s;
  nsqrt = 1.0 + sqrt(1.0 + (1.0 + s) / s * 2.0 / n0 / dm);
  n1 = (int)(n0 / 4.0 * nsqrt * nsqrt + 0.99);
  n2 = (int)((double)n1 * s + 0.99);
  fout << "  検出力=" << opt.po() << "・検出差=" << dm << "の時の1群必要例数=" << n1 << "(2群合計必要例数=" << n1 + n2 << ")" << endl;
}

void ctestclass::put_rr(ostream& fout, option& opt, const ftable& ftab) const
// リスク比または陽性尤度比を出力する
{
  double rr = 0.0, b = 0.0, v = 0.0, csqr = 0.0;
  if (cal_rr(rr, b, v, csqr, opt, ftab) == false)
  {
    if (opt.tab() == tabfprc) fout << "\n・陽性尤度比(positive likelihood ratio)";
    else fout << "\n・リスク比(risk ratio)";
    fout << ":計算不可能!\n";
    return;
  }
  ldbl pr = ctop(csqr, 1);
  if (opt.tab() == tabfprc) fout << "\n・陽性尤度比(positive likelihood ratio):LR=" << rr << "  ln(LR)=" << b << "\n";
  else fout << "\n・リスク比(risk ratio):RR=" << rr << "  ln(RR)=" << b << "\n";
  double ci = ptoz(1.0 - opt.cc()) * sqrt(v);
  double bl = b - ci, bu = b + ci;
  if (opt.tab() == tabfprc) fout << "  陽性尤度比の";
  else fout << "  リスク比の";
  fout << opt.cc() * 100.0 << "%信頼区間:下限=" << exp(bl) << "  上限=" << exp(bu) << "\n";
  fout << "  χ2乗=" << csqr << "  自由度=1  有意確率:p=" << pr << pind(pr) << endl;
}

bool ctestclass::cal_rr(double& rr, double& b, double& v, double& csqr, option& opt, const ftable& ftab) const
// リスク比または陽性尤度比とその検定を行う
{
  double n12 = (double)(ftab.f(1, 2)), nt1 = (double)(ftab.frtotal(1));
  double n22 = (double)(ftab.f(2, 2)), nt2 = (double)(ftab.frtotal(2));
  if (opt.tab() == tabfprc)
  {
    n12 = (double)(ftab.f(2, 1));  nt1 = (double)(ftab.fctotal(1));
    n22 = (double)(ftab.f(2, 2));  nt2 = (double)(ftab.fctotal(2));
  }
  rr = b = v = csqr = 0.0;
  if (n12 <= 0.0) return false; // 計算不可能
  if (n22 > 0.0)
  {
    rr = n22 / nt2 * nt1 / n12;  b = log(rr);
    v = 1.0 / n22 - 1.0 / nt2 + 1.0 / n12 - 1.0 / nt1;
    csqr = (sgn(v) == 0) ? 0.0 : b * b / v;
  }
  return true;                  // 計算可能
}

void ctestclass::put_odr(ostream& fout, option& opt, const ftable& ftab) const
// オッズ比を出力する
{
  double odr = 0.0, b = 0, v = 0.0, csqr = 0.0;
  testtype odrtype = cal_odr(odr, b, v, csqr, opt, ftab);
  if (odrtype == testnon)
  {
    fout << "\n・オッズ比(odds ratio):計算不可能!\n";
    return;
  }
  ldbl pr = ctop(csqr, 1);
  fout << "\n・オッズ比(odds ratio";
  if (odrtype == testcor) fout << ",Woolfの修正有";
  fout << "):OR=" << odr << "  ln(OR)=" << b << "\n";
  double ci = ptoz(1.0 - opt.cc()) * sqrt(v);
  double bl = b - ci, bu = b + ci;
  fout << "  オッズ比の" << opt.cc() * 100.0 << "%信頼区間:下限=" << exp(bl) << "  上限=" << exp(bu) << "\n";
  fout << "  χ2乗=" << csqr << "  自由度=1  有意確率:p=" << pr << pind(pr) << endl;
}

testtype ctestclass::cal_odr(double& odr, double& b, double& v, double& csqr, option& opt, const ftable& ftab) const
// オッズ比とその検定を行う
{
  double n11 = (double)(ftab.f(1, 1)), n12 = (double)(ftab.f(1, 2));
  double n21 = (double)(ftab.f(2, 1)), n22 = (double)(ftab.f(2, 2));
  bool woolf = false;
  odr = b = v = csqr = 0.0;
  if (ftab.f(1, 1) == 0 || ftab.f(1, 2) == 0 || ftab.f(2, 1) == 0 || ftab.f(2, 2) == 0)
  {
    if (opt.odr() != testaut) return testnon; // 計算不可能
    n11 += 0.5;  n12 += 0.5;  n21 += 0.5;  n22 += 0.5;
    woolf = true;               // Woolfの修正有
  }
  odr = n11 / n12 * n22 / n21;
  b = log(odr);
  v = 1.0 / n11 + 1.0 / n12 + 1.0 / n21 + 1.0 / n22;
  csqr = (sgn(v) == 0) ? 0.0 : b * b / v;
  return woolf ? testcor : testorg;
}

void ctestclass::put_st(ostream& fout, option& opt, const ftable& ftab) const
// 交差度数の優劣比較の結果を出力する
{
  int f1 = ftab.f(1,2), f2 = ftab.f(2,1);
  int n = f1 + f2, dif = f1 - f2;
  double pr1 = (double)f1 / (double)n, pr2 = (double)f2 / (double)n;
  fout << "\n・交差度数の優劣比較=" << f1 << "(" << 100.0 * pr1 << "%):" << f2 << "(" << 100.0 * pr2 << "%)\n";
  put_stpci(fout, opt, f1, f2);
  ldbl pr0 = binom(f1, f2);
  ldbl pr = (pr0 < 0.5L) ? 2.0L * pr0 : 1.0L;
  fout << "  符号検定 有意確率:片側p=" << pr0 << pind(pr0) << " 両側p=" << pr << pind(pr) << "\n";
  double z = (double)(dif - sgn((double)dif)) / sqrt((double)n);
  pr = ztop(z);
  fout << "  McNemarの検定(連続修正有):z=" << z << "  両側有意確率:p=" << pr << pind(pr) << endl;
}

void ctestclass::put_stpci(ostream& fout, option& opt, int f1, int f2) const
// 交差度数の出現率の信頼区間出力
{
  int n = f1 + f2, dif = f1 - f2;
  double p0 = (double)f1 / (double)n;
  double plow = 0.0, pup = 1.0;
  double pci = calbexci(plow, pup, f1, n, opt.cc());
  fout << "  交差度数" << f1 << "の出現率と理論確率の差=" << p0 << "-0.5=" << p0 - 0.5 << "\n";
  fout << "  差の" << opt.cc() * 100.0 << "%信頼区間:下限=" << plow - 0.5 << "  上限=" << pup - 0.5 << endl;
}

void ctestclass::put_mcon(ostream& fout, option& opt, const ftable& ftab) const
// 群間の多重比較結果を出力する
{
  if (opt.res("ma")) put_macon(fout, opt.ma(), opt, ftab);
  if (opt.res("md")) put_mdcon(fout, opt.md(), opt, ftab);
  if (ftab.nc() > 2) return;    // 3分類以上はここまで
  if (opt.res("rr"))            // リスク比
  {
    if (opt.res("ma"))
    {
      put_rrmcon(fout, opt.ma(), true, opt, ftab);
      put_rrmconci(fout, opt.ma(), true, opt, ftab);
    }
    if (opt.res("md"))
    {
      put_rrmcon(fout, opt.md(), false, opt, ftab);
      put_rrmconci(fout, opt.md(), false, opt, ftab);
    }
  }
  if (opt.res("or"))            // オッズ比
  {
    if (opt.res("ma"))
    {
      put_odrmcon(fout, opt.ma(), true, opt, ftab);
      put_odrmconci(fout, opt.ma(), true, opt, ftab);
    }
    if (opt.res("md"))
    {
      put_odrmcon(fout, opt.md(), false, opt, ftab);
      put_odrmconci(fout, opt.md(), false, opt, ftab);
    }
  }
}

void ctestclass::put_macon(ostream& fout, mctype mc, option& opt, const ftable& ftab) const
// 全ての2群間の多重比較結果を出力する
{
  if (mc == mcetc)
  {
    if (ftab.nc() > 2)
    {
      put_cmcon(fout, mcfis, true, opt, ftab);
      put_cmcon(fout, mctuk, true, opt, ftab);
      put_cmcon(fout, mcbon, true, opt, ftab);
      put_cmcon(fout, mcsch, true, opt, ftab);
    }
    else
    {
      put_cmcon(fout, mcfis, true, opt, ftab);
      put_cmconci(fout, mcfis, true, opt, ftab);
      if (opt.res("pa")) put_mpa(fout, mcfis, true, opt, ftab);
      put_cmcon(fout, mctuk, true, opt, ftab);
      put_cmconci(fout, mctuk, true, opt, ftab);
      put_cmcon(fout, mcbon, true, opt, ftab);
      put_cmconci(fout, mcbon, true, opt, ftab);
      if (opt.res("pa")) put_mpa(fout, mcbon, true, opt, ftab);
      put_cmcon(fout, mcsch, true, opt, ftab);
      put_cmconci(fout, mcsch, true, opt, ftab);
      if (opt.res("pa")) put_mpa(fout, mcsch, true, opt, ftab);
    }
  }
  else
  {
    put_cmcon(fout, mc, true, opt, ftab);
    if (ftab.nc() == 2)
    {
      put_cmconci(fout, mc, true, opt, ftab);
      if (opt.res("pa")) put_mpa(fout, mc, true, opt, ftab);
    }
  }
}

void ctestclass::put_mdcon(ostream& fout, mctype mc, option& opt, const ftable& ftab) const
// 対照群と他の群の多重比較結果を出力する
{
  if (mc == mcetc)
  {
    if (ftab.nc() > 2)
    {
      put_cmcon(fout, mcfis, false, opt, ftab);
      put_cmcon(fout, mcdun, false, opt, ftab);
      put_cmcon(fout, mcbon, false, opt, ftab);
      put_cmcon(fout, mcsch, false, opt, ftab);
    }
    else
    {
      put_cmcon(fout, mcfis, false, opt, ftab);
      put_cmconci(fout, mcfis, false, opt, ftab);
      put_cmcon(fout, mcdun, false, opt, ftab);
      put_cmconci(fout, mcdun, false, opt, ftab);
      if (opt.res("pa")) put_mpa(fout, mcbon, false, opt, ftab);
      put_cmcon(fout, mcbon, false, opt, ftab);
      put_cmconci(fout, mcbon, false, opt, ftab);
      put_cmcon(fout, mcsch, false, opt, ftab);
      put_cmconci(fout, mcsch, false, opt, ftab);
      if (opt.res("pa")) put_mpa(fout, mcsch, false, opt, ftab);
    }
  }
  else
  {
    put_cmcon(fout, mc, false, opt, ftab);
    if (ftab.nc() == 2)
    {
      put_cmconci(fout, mc, false, opt, ftab);
      if (opt.res("pa")) put_mpa(fout, mc, false, opt, ftab);
    }
  }
}

void ctestclass::put_bun(ostream& fout, option& opt, idd& id, const ftable& ftab) const
// m×2累積分割表を出力する
{
  vector<int> fcumvec(ftab.nr());
  for (size_t i = 0; i < ftab.nr(); ++i) fcumvec[i] = 0;
  for (size_t j = 1; j < ftab.nc(); ++j)
  {
    ftable ftabj(ftab);
    ftabj.resize(ftab.nr(), 2);
    for (size_t i = 1; i <= ftab.nr(); ++i)
    {
      fcumvec[i - 1] += ftab.f(i, j);
      ftabj.assign(i, 1, fcumvec[i - 1]);
      ftabj.assign(i, 2, ftab.frtotal(i) - fcumvec[i - 1]);
    }
    ftabj.ccode(1, "["+ftab.ccode(1)+"]-["+ftab.ccode(j)+"]");
    ftabj.ccode(2, "["+ftab.ccode(j + 1)+"]-["+ftab.ccode(ftab.nc())+"]");
    ftabj.set();
    ftabj.trm();              // 行・列の合計が0のものを削除する
    if (ftabj.nc() > 1)
    {
      put_id(fout, id);
      if (opt.res("tab")) putftable(fout, ftabj, opt.tab());
      if (opt.res("r")) put_cramer(fout, opt, ftabj);
      if (opt.res("lt")) put_ltest(fout, ftabj);
      if (ftabj.nr() == 2)
      {
        if (opt.res("rd")) put_rd(fout, opt, ftabj);
        if (opt.res("rr")) put_rr(fout, opt, ftabj);
        if (opt.res("or")) put_odr(fout, opt, ftabj);
      }
      if (ftab.nr() > 2)
      {
        if (opt.res("ma") || opt.res("md")) put_mcon(fout, opt, ftabj);
      }
      fout << "\n" << endl;
    }
  }
}

void ctestclass::put_bun2(ostream& fout, option& opt, idd& id, const ftable& ftab) const
// m×2単独分割表を出力する
{
  size_t jend = ftab.nc();
  if (opt.b2() == 1) jend = 2;
  for (size_t j = 1; j < jend; ++j)
  {
    for (size_t k = j + 1; k <= ftab.nc(); ++k)
    {
      ftable ftabj(ftab);
      ftabj.resize(ftab.nr(), 2);
      for (size_t i = 1; i <= ftab.nr(); ++i)
      {
        ftabj.assign(i, 1, ftab.f(i, j));
        ftabj.assign(i, 2, ftab.f(i, k));
      }
      ftabj.ccode(1, ftab.ccode(j));
      ftabj.ccode(2, ftab.ccode(k));
      ftabj.set();
      ftabj.trm();              // 行・列の合計が0のものを削除する
      if (ftabj.nr() > 1 && ftabj.nc() > 1)
      {
        put_id(fout, id);
        if (opt.res("tab")) putftable(fout, ftabj, opt.tab());
        if (opt.res("r")) put_cramer(fout, opt, ftabj);
        if (opt.res("lt")) put_ltest(fout, ftabj);
        if (ftabj.nr() == 2)
        {
          if (opt.res("rd")) put_rd(fout, opt, ftabj);
          if (opt.res("rr")) put_rr(fout, opt, ftabj);
          if (opt.res("or")) put_odr(fout, opt, ftabj);
        }
        if (ftab.nr() > 2)
        {
          if (opt.res("ma") || opt.res("md")) put_mcon(fout, opt, ftabj);
        }
        fout << "\n" << endl;
      }
    }
  }
}

void ctestclass::put_cmcon(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const
// 出現率の差の多重比較結果を出力する
{
  int df = ftab.nc() - 1, dc = 1;
  string method("Tukey");
  string tstr("出現率の差");
  if (ftab.nc() > 2)
  {
    tstr = "群";
    if (mc == mcdun || mc == mctuk) mc = mcbon; // 分類数が3以上はBonferroni型
  }
  switch (mc)
  {
  case mcbon : method = "Bonferroni";
    dc = ma ? ftab.nr() * (ftab.nr() - 1) / 2 : ftab.nr() - 1;
    break;
  case mcdun : method = "Dunnett";
    break;
  case mcfis : method = "Fisher";
    break;
  case mcsch : method = "Scheffe";
    df = (ftab.nr() - 1) * (ftab.nc() - 1);
  }
  fout << "\n・" << tstr << "の" << method << "型多重比較";
  if (ftab.nc() == 2 && opt.res("ym")) fout << "(連続修正有)";
  fout << "\n";
  int llen = 65;                // 区切り線の長さ
  fout << "群         - 群        ";
  if (mc == mcdun) fout << "       d値  群数";
  else if (mc == mctuk) fout << "       q値  群数";
  else { fout << "      χ2乗値";  llen -= 4; }
  fout << "  自由度   有意確率p値\n";
  putline(fout, llen);
  int nt = ftab.ftotal();
  size_t iend = ma ? ftab.nr() : 2;
  for (size_t i = 1; i < iend; ++i)
  {
    double xi = (double)(ftab.frtotal(i));
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      double xj = (double)(ftab.frtotal(j)), csqr = 0.0;
      double cor = 0.5 * (1.0 / xi + 1.0 / xj);
      for (size_t k = 1; k <= ftab.nc(); ++k)
      {
        double pd = (double)(ftab.f(j, k)) / xj - (double)(ftab.f(i, k)) / xi;
        if (ftab.nc() == 2 && opt.res("ym")) pd -= sgn(pd) * cor; // Yatesの修正
        csqr += pd * pd / (double)(ftab.fctotal(k));
      }
      csqr *= nt * xi * xj / (xi + xj);
      double r = (xi == xj) ? 0.5 : sqrt(xi / (xi + xj) * xj / (xi + xj));
      ldbl pr = calmcpr(mc, csqr, ftab.nr(), df, 0, dc, r);
      int npad = ftab.rcode(j).size() - strclen(ftab.rcode(j), CHARwide);
      fout << left << setw(11 + npad) << ftab.rcode(j);
      fout << "- " << setw(10 + npadi) << codei;
      if (mc == mcdun || mc == mctuk) fout << right << setw(10) << csqr << setw(6) << ftab.nr() << setw(8) << "inf";
      else fout << right << setw(12) << csqr << setw(8) << df;
      fout << setw(14) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, llen);
}

void ctestclass::put_cmconci(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const
// 出現率の差の同時信頼区間を出力する
{
  int df = 1, dc = 1;
  string method("Tukey");
  switch (mc)
  {
  case mcbon : method = "Bonferroni";
    dc = ma ? ftab.nr() * (ftab.nr() - 1) / 2 : ftab.nr() - 1;
    break;
  case mcdun : method = "Dunnett";
    break;
  case mcfis : method = "Fisher";
    break;
  case mcsch : method = "Scheffe";
    df = ftab.nr() - 1;
  }
  bool ym = opt.res("ym");
  fout << "\n・出現率の差の" << method << "型" << opt.cc() * 100.0 << "%同時信頼区間";
  if (ftab.nc() == 2 && ym) fout << "(連続修正有)";
  fout << "\n";
  const int LENline = 71;       // 区切り線の長さ
  fout << "群         - 群          出現率の差    絶対精度        下限        上限\n";
  putline(fout, LENline);
  int nt = ftab.ftotal();
  size_t iend = ma ? ftab.nr() : 2;
  for (size_t i = 1; i < iend; ++i)
  {
    double xi = (double)(ftab.frtotal(i));
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      double xj = (double)(ftab.frtotal(j));
      int npad = ftab.rcode(j).size() - strclen(ftab.rcode(j), CHARwide);
      fout << left << setw(11 + npad) << ftab.rcode(j);
      fout << "- " << setw(10 + npadi) << codei;
      double r = (xi == xj) ? 0.5 : sqrt(xi / (xi + xj) * xj / (xi + xj));
      double d = 0.0;
      double pci = caldpci(d, mc, opt.cc(), ftab.nr(), df, dc, ftab.f(i, 1), ftab.f(i, 2), ftab.f(j, 1), ftab.f(j, 2), ym, r);
      fout << right << setw(12) << d << setw(12) << pci << setw(12) << d - pci << setw(12) << d + pci << "\n";
    }
  }
  putline(fout, LENline);
}

void ctestclass::put_mpa(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const
// 出現率の差の検出力分析結果を出力する
{
  string method("Fisher");
  int df = 1;
  double dc = 1.0;
  double ac = opt.al(), poc = opt.po();
  double bc = 1.0 - poc, zac = ptoz(ac);
  double z2bc = (poc > 0.5) ? ptoz(2.0 * bc) : -ptoz(2.0 * poc);
  switch (mc)
  {
  case mcdun :
  case mctuk :
  case mcbon : method = "Bonferroni";
    dc = ma ? (double)(ftab.nr() * (ftab.nr() - 1) / 2) : (double)(ftab.nr() - 1);
    ac = ac / dc;
    zac = ptoz(ac);
    break;
  case mcsch : method = "Scheffe";
    df = ftab.nr() - 1;
    zac = sqrt(ptoc(ac, df));
  }
  fout << "\n・検出力分析:設定有意水準=" << opt.al() << "  設定検出力=" << opt.po() << "  " << method << "型多重比較\n";
  const int LENline = 71;       // 区切り線の長さ
  fout << "群         - 群          設定検出差  調整検出力  調整検出差 1群必要例数\n";
  putline(fout, LENline);
  int nt = ftab.ftotal();
  ldbl bcc = 1.0L;
  size_t iend = ma ? ftab.nr() : 1;
  for (size_t i = 1; i <= iend; ++i)
  {
    string code1(ftab.rcode(i));
    int npad1 = code1.size() - strclen(code1, CHARwide);
    double n1 = (double)(ftab.frtotal(i)), n12 = (double)(ftab.f(i, 2));
    double p1 = n12 / n1;
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      int npad = ftab.rcode(j).size() - strclen(ftab.rcode(j), CHARwide);
      fout << left << setw(11 + npad) << ftab.rcode(j);
      fout << "- " << setw(10 + npad1) << code1;
      double n2 = (double)(ftab.frtotal(j)), n22 = (double)(ftab.f(j, 2));
      double p2 = n22 / n2;
      double dp = fabs(p1 - p2);
      double dm = dp;
      if (opt.dm() > 0.0 || sgn(dm) == 0)
      {
        dm = opt.dm();
        if (dm <= 0.0) dm = 0.1; // 出現率の差が0の時は0.1にする
        p2 = (p1 < 1.0) ? p1 + dm : 1.0 - dm;
        if (p2 > 1.0) p2 = 1.0;
      }
      double r = n2 / n1;
      double pi = (p1 + r * p2) / (r + 1.0);
      double v = pi * (1.0 - pi);
      double tmp1 = zac * sqrt((r + 1.0) * v);
      double tmp2 = sqrt(r * p1 * (1.0 - p1) + p2 * (1.0 - p2));
      double n1t = n1 * 0.9;
      double tmp3 = 1.0 + sqrt(1.0 + 2.0 * (r + 1.0) / r / n1t / dm);
      n1t = 4.0 * n1 / tmp3 / tmp3;
      double z2b = (sgn(tmp2) == 0) ? 0.0 : (sqrt(r * dm * dm * n1t) - tmp1) / tmp2;
      ldbl po = sndis(z2b);
      ldbl b = 1.0L - po;
      bcc *= b;
      double dpc = (zac + z2bc) * sqrt((1.0 + r) / r * v / n1);
      double nsqrt = (zac * sqrt(2.0 * v) + z2bc * sqrt(p1 * (1.0 - p1) + p2 * (1.0 - p2))) / dm;
      double n0 = nsqrt * nsqrt;
      nsqrt = 1.0 + sqrt(1.0 + 4.0 / n0 / dm);
      n0 = (int)(n0 / 4.0 * nsqrt * nsqrt + 0.99);
      fout << right << setw(12) << dm << setw(12) << po << setw(12) << dpc << setw(12) << n0 << "\n";
    }
  }
  putline(fout, LENline);
  if (mc != mcfis) fout << "全体の調整検出力(連続修正有)=" << 1.0 - bcc << "(β=" << bcc << ")" << endl;
}

void ctestclass::put_rrmcon(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const
// リスク比の多重比較結果を出力する
{
  int df = 1;
  int dc = ma ? ftab.nr() * (ftab.nr() - 1) / 2 : ftab.nr() - 1;
  string method("Bonferroni");
  if (mc == mcfis) { method = "Fisher";  dc = 1; }
  else mc = mcbon;              // Fisher以外はBonferroniにする
  fout << "\n・リスク比の" << method << "型多重比較\n";
  int llen = 61;                // 区切り線の長さ
  fout << "群         - 群              χ2乗値  自由度   有意確率p値\n";
  putline(fout, llen);
  ftable ftabi(2, 2);
  size_t iend = ma ? ftab.nr() : 2;
  for (size_t i = 1; i < iend; ++i)
  {
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      ftabi.assign(1, 1, ftab.f(i, 1));  ftabi.assign(1, 2, ftab.f(i, 2));
      ftabi.assign(2, 1, ftab.f(j, 1));  ftabi.assign(2, 2, ftab.f(j, 2));
      double rr = 0.0, b = 0.0, v = 0.0, csqr = 0.0;
      int npad = ftab.rcode(j).size() - strclen(ftab.rcode(j), CHARwide);
      fout << left << setw(11 + npad) << ftab.rcode(j);
      fout << "- " << setw(10 + npadi) << codei;
      if (cal_rr(rr, b, v, csqr, opt, ftabi) == false)
      { fout << right << setw(12) << "-" << "\n"; }
      else
      {
        double pr = calmcpr(mc, csqr, ftab.nr(), df, 0, dc);
        fout << right << setw(12) << csqr << setw(8) << df;
        fout << setw(14) << pr << pind(pr) << "\n";
      }
    }
  }
  putline(fout, llen);
}

void ctestclass::put_rrmconci(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const
// リスク比の同時信頼区間を出力する
{
  int df = 1;
  int dc = ma ? ftab.nr() * (ftab.nr() - 1) / 2 : ftab.nr() - 1;
  string method("Bonferroni");
  if (mc == mcfis) { method = "Fisher";  dc = 1; }
  else mc = mcbon;              // Fisher以外はBonferroniにする
  fout << "\n・リスク比の" << method << "型" << opt.cc() * 100.0 << "%同時信頼区間\n";
  const int LENline = 71;       // 区切り線の長さ
  fout << "群         - 群        対数リスク比    リスク比        下限        上限\n";
  putline(fout, LENline);
  ftable ftabi(2, 2);
  size_t iend = ma ? ftab.nr() : 2;
  for (size_t i = 1; i < iend; ++i)
  {
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      ftabi.assign(1, 1, ftab.f(i, 1));  ftabi.assign(1, 2, ftab.f(i, 2));
      ftabi.assign(2, 1, ftab.f(j, 1));  ftabi.assign(2, 2, ftab.f(j, 2));
      double rr = 0.0, b = 0.0, v = 0.0, csqr = 0.0;
      int npad = ftab.rcode(j).size() - strclen(ftab.rcode(j), CHARwide);
      fout << left << setw(11 + npad) << ftab.rcode(j);
      fout << "- " << setw(10 + npadi) << codei;
      if (cal_rr(rr, b, v, csqr, opt, ftabi) == false)
      { fout << right << setw(12) << "-" << "\n"; }
      else
      {
        double pr = opt.cc() / (double)dc;
        double ci = ptoz(1.0 - pr) * sqrt(v);
        double bl = b - ci, bu = b + ci;
        fout << right << setw(12) << b << setw(12) << rr << setw(12) << exp(bl) << setw(12) << exp(bu) << "\n";
      }
    }
  }
  putline(fout, LENline);
}

void ctestclass::put_odrmcon(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const
// オッズ比の多重比較結果を出力する
{
  int df = 1;
  int dc = ma ? ftab.nr() * (ftab.nr() - 1) / 2 : ftab.nr() - 1;
  string method("Bonferroni");
  if (mc == mcfis) { method = "Fisher";  dc = 1; }
  else mc = mcbon;              // Fisher以外はBonferroniにする
  fout << "\n・オッズ比の" << method << "型多重比較\n";
  int llen = 61;                // 区切り線の長さ
  fout << "群         - 群              χ2乗値  自由度   有意確率p値\n";
  putline(fout, llen);
  ftable ftabi(2, 2);
  size_t iend = ma ? ftab.nr() : 2;
  for (size_t i = 1; i < iend; ++i)
  {
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      ftabi.assign(1, 1, ftab.f(i, 1));  ftabi.assign(1, 2, ftab.f(i, 2));
      ftabi.assign(2, 1, ftab.f(j, 1));  ftabi.assign(2, 2, ftab.f(j, 2));
      double odr = 0.0, b = 0.0, v = 0.0, csqr = 0.0;
      int npad = ftab.rcode(j).size() - strclen(ftab.rcode(j), CHARwide);
      fout << left << setw(11 + npad) << ftab.rcode(j);
      fout << "- " << setw(10 + npadi) << codei;
      if (cal_odr(odr, b, v, csqr, opt, ftabi) == testnon)
      { fout << right << setw(12) << "-" << "\n"; }
      else
      {
        ldbl pr = ctop(csqr, 1) * (ldbl)(dc);
        if (pr > 1.0) pr = 1.0;
        fout << right << setw(12) << csqr << setw(8) << df;
        fout << setw(14) << pr << pind(pr) << "\n";
      }
    }
  }
  putline(fout, llen);
}

void ctestclass::put_odrmconci(ostream& fout, mctype mc, bool ma, option& opt, const ftable& ftab) const
// オッズ比の同時信頼区間を出力する
{
  int df = 1;
  int dc = ma ? ftab.nr() * (ftab.nr() - 1) / 2 : ftab.nr() - 1;
  string method("Bonferroni");
  if (mc == mcfis) { method = "Fisher";  dc = 1; }
  else mc = mcbon;              // Fisher以外はBonferroniにする
  fout << "\n・オッズ比の" << method << "型" << opt.cc() * 100.0 << "%同時信頼区間\n";
  const int LENline = 71;       // 区切り線の長さ
  fout << "群         - 群        対数オッズ比    オッズ比        下限        上限\n";
  putline(fout, LENline);
  ftable ftabi(2, 2);
  size_t iend = ma ? ftab.nr() : 2;
  for (size_t i = 1; i < iend; ++i)
  {
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      ftabi.assign(1, 1, ftab.f(i, 1));  ftabi.assign(1, 2, ftab.f(i, 2));
      ftabi.assign(2, 1, ftab.f(j, 1));  ftabi.assign(2, 2, ftab.f(j, 2));
      double odr = 0.0, b = 0.0, v = 0.0, csqr = 0.0;
      int npad = ftab.rcode(j).size() - strclen(ftab.rcode(j), CHARwide);
      fout << left << setw(11 + npad) << ftab.rcode(j);
      fout << "- " << setw(10 + npadi) << codei;
      if (cal_odr(odr, b, v, csqr, opt, ftabi) == testnon)
      { fout << right << setw(12) << "-" << "\n"; }
      else
      {
        double pr = opt.cc() / (double)dc;
        double ci = ptoz(1.0 - pr) * sqrt(v);
        double bl = b - ci, bu = b + ci;
        fout << right << setw(12) << b << setw(12) << odr << setw(12) << exp(bl) << setw(12) << exp(bu) << "\n";
      }
    }
  }
  putline(fout, LENline);
}

int ctestclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("実行時エラー！");
  if (opt.check(nftab) != 0) throw option_error("実行時オプションエラー！");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void ctestclass::put(option& opt, listidd& idlst, listftab& ftablst) const
// 結果を出力する
{
  if (opt.res("res") == false) return;
  bool putendl = true;          // 最後の2行の出力フラッグ
  listidd_itr iitr = idlst.begin();
  for (auto& ftabi : ftablst)
  {
    put_id(cout, *iitr);
    ftabi.trm();                // 行・列の合計が0のものを削除する
    if (ftabi.nr() > 0)
    {
      if (opt.res("tab")) putftable(cout, ftabi, opt.tab());
      if (ftabi.nr() > 1 && ftabi.nc() > 1)
      {
        if (opt.res("def") || opt.res("r"))
        {
          if (opt.res("r") || (opt.res("def") && (ftabi.nr() > 2 || ftabi.nc() > 2))) put_cramer(cout, opt, ftabi);
        }
        if (opt.res("lt")) put_ltest(cout, ftabi);
        if (ftabi.nr() == 2 && ftabi.nc() == 2)
        {
          if (opt.res("rd")) put_rd(cout, opt, ftabi);
          if (opt.res("pa")) put_pa(cout, opt, ftabi);
          if (opt.res("rr")) put_rr(cout, opt, ftabi);
          if (opt.res("or")) put_odr(cout, opt, ftabi);
          if (opt.res("st")) put_st(cout, opt, ftabi);
        }
        if (ftabi.nr() > 2 && (opt.res("ma") || opt.res("md")))
        {
          put_mcon(cout, opt, ftabi);
        }
        if (ftabi.nc() > 2 && (opt.res("b2") || opt.res("bc")))
        {
          cout << "\n" << endl;
          if (opt.res("bc")) put_bun(cout, opt, *iitr, ftabi);
          if (opt.res("b2")) put_bun2(cout, opt, *iitr, ftabi);
          putendl = false;
        }
      }
    }
    if (putendl) cout << "\n" << endl;
    else putendl = true;
    ++iitr;
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  ctestclass ctest;             // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          ctest.get(nullptr, opt, idlst, ftablst);
          ctest.put(opt, idlst, ftablst);
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        ctest.get(argv[i], opt, idlst, ftablst);  fok = false;
        ctest.put(opt, idlst, ftablst);
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
