// htest.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2024-12-16 09:43:57 tac>
//
// クロス集計表を入力し、Kruskal-Wallisの検定とU検定を行う。
// クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W htest htest.C
//              -I/usr/local/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "htest  Version 7.42          Copyleft(Vo\\oV) 2024 杉本典夫",
  "書式:htest  [<オプション> …]  [<クロス集計表ファイル> …]",
  "機能:クロス集計表に対してKruskal-Wallisの検定(U検定含む)をするよ。",
  "<オプション>",
  "  -cc[=<信頼係数>]:信頼係数を指定するよ。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込むよ。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するよ。(-h:-h=man)",
  "  -res=[-][{def|all|bt[{b|f}]|ht|m[{a|d}][{b|d|f|s|t}]|r",
  "       tab[{fr|pc|pr|pt}]|ut[{a|c|o}]}][,…]:解析結果の出力項目を指定するよ。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    bt[{b|f}]:2×2分割表の多重比較(b:Bonferroni型 f:Fisher型)",
  "    ht:Kruskal-Wallisの検定(H検定)",
  "    m[{a|d}][{b|d|f|s|t}]:多重比較(ma:全ての2群比較 md:群1と他の群の比較",
  "      b:Bonferroni型 d:Dunnett型 f:Fisher型 s:Scheffe型 t:Tukey型)",
  "    r:順位相関比",
  "    tab[={fr|pc|pr|pt}]:クロス集計表(fr:度数だけ pc:列合計に対する％出力",
  "      pr:行合計に対する％出力 pt:全合計に対する％出力)",
  "    ut[{a|c|o}]:Mann-WhitneyのU検定",
  "      a:連続修正自動判定  c:連続修正有  o:連続修正無",
  "    オプション省略時:データに合わせて自動判定"
};

const string Title = "=== 順序データの群間比較 ==="; // 表題

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  mctype vbt;                   // 2×2分割表の多重比較法
  double vcc;                   // 信頼係数
  mctype vma;                   // 全ての2群ごとの多重比較法
  mctype vmd;                   // 対照群と他の群の多重比較法
  lstr vreslst;                 // 結果出力リスト
  mapsb vresmap;                // 結果出力の有無マップ
  tabtype vtab;                 // 度数集計表の種類
  testtype vut;                 // U検定の種類
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  mctype bt() const             // 2×2分割表の多重比較法を返す
  { return vbt; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  mctype ma() const             // 全ての2群ごとの多重比較法を返す
  { return vma; }
  mctype md() const             // 対照群と他の群の多重比較法を返す
  { return vmd; }
  bool res(const string& code); // 解析結果出力フラッグを返す
  tabtype tab() const           // 度数集計表出力フラッグを返す
  { return vtab; }
  testtype ut() const           // U検定の種類を返す
  { return vut; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int nfile);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vbt(mcbon), vcc(CCdef), vma(mctuk), vmd(mcdun), vreslst(), vresmap(), vtab(tabfprr), vut(testaut)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 結果出力フラッグを返す
{
  mapsb_citr respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(int nftab)    // オプションのセットとチェック
{
  if (nftab < 1) return disperr(vpname, "クロス集計表がない！");
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["tab"] = true;
  vresmap["bt"] = vresmap["ht"] = vresmap["ma"] = vresmap["md"] = vresmap["r"] = vresmap["ut"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["tab"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr respos = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(resstr.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def")    // 省略時出力
      { vresmap["def"] = vresmap["tab"] = true; }
      else if (resstr == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["bt"] = vresmap["ht"] = vresmap["ma"] = vresmap["md"] = vresmap["r"] = vresmap["tab"] = vresmap["ut"] = true;
        vbt = mcbon;;  vma = mctuk;  vmd = mcdun;  vtab = tabfprr;  vut = testaut;
      }
      else if (resstr == "btb") // Bonferroni型
      { vresmap["bt"] = true;  vbt = mcbon; }
      else if (resstr == "btf") // Fisher型
      { vresmap["bt"] = true;  vbt= mcfis; }
      else if (resstr == "m")      // -res=mと指定された時
      { vresmap["ma"] = vresmap["md"] = true; }
      else if (resstr[0] == 'm' && resstr[1] == 'a' && resstr.size() > 2)
      {
        switch (resstr[2])
        {
        case 'b' : vma = mcbon;
          break;
        case 'f' : vma = mcfis;
          break;
        case 's' : vma = mcsch;
          break;
        case 't' : vma = mctuk;
          break;
        default : return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！");
        }
        vresmap["ma"] = true;
      }
      else if (resstr[0] == 'm' && resstr[1] == 'd' && resstr.size() > 2)
      {
        switch (resstr[2])
        {
        case 'b' : vmd = mcbon;
          break;
        case 'd' : vmd = mcdun;
          break;
        case 'f' : vmd = mcfis;
          break;
        case 's' : vmd = mcsch;
          break;
        default : return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！");
        }
        vresmap["md"] = true;
      }
      else if (resstr == "tabfr") // 度数だけ出力
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (resstr == "tabpc") // 列合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (resstr == "tabpr") // 行合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (resstr == "tabpt") // 全合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (resstr == "uta")   // U検定(修正自動判定)
      { vresmap["ut"] = true;  vut= testaut; }
      else if (resstr == "utc")   // U検定(修正有り)
      { vresmap["ut"] = true;  vut= testcor; }
      else if (resstr == "uto")   // U検定(修正無し)
      { vresmap["ut"] = true;  vut = testorg; }
      else if (respos == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

class htestclass                // htestメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_rr(ostream& fout, option& opt, int df, double csqr, const ftable& ftab) const;
  // 順位相関比を出力する
  void put_utest(ostream& fout, option& opt, vecdbl& tvec, const ftable& ftab) const;
  // U検定の結果を出力する
  void put_bun(ostream& fout, option& opt, const ftable& ftab) const;
  // 2×2分割表に関する多重比較結果を出力する
  void put_hmcon(ostream& fout, mctype mc, bool ma, double kc, const vecdbl& tvec, const ftable& ftab) const;
  // 順位平均値の多重比較を出力する
  void put_hmrdci(ostream& fout, option& opt, mctype mc, bool ma, double kc, const vecdbl& tvec, const ftable& ftab) const;
  // 2群間の順位平均値の差の信頼区間出力
public:
  htestclass() : fname()        // コンストラクタ
  {}
  ~htestclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // 度数集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst) const;
  // 結果を出力する
};

void htestclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "群項目(縦)  :" << id.item(i) << "\n";
      break;
    case avalcol : fout << "順序項目(横):" << id.item(i) << "\n";
      break;
    default : id.put_item(fout, i);
    }
  }
}

void htestclass::put_rr(ostream& fout, option& opt, int df, double csqr, const ftable& ftab) const
// 順位相関比を出力する
{
  int dft = ftab.ftotal() - 1;
  double r2 = csqr / (double)dft;
  fout << "\n・順位相関比:ηr=" << sqrt(r2)<< "  寄与率:ηr2乗=" << r2 << "\n";
  double csqrl = 0.0, csqru = 0.0;
  calcsqrci(csqrl, csqru, opt.cc(), df, csqr);
  double rl = sqrt(csqrl / (double)dft);
  double ru = sqrt(csqru / (double)dft);
  if (ru > 1.0) ru = 1.0;
  fout << "  順位相関比の" << opt.cc() * 100.0 << "%信頼区間:下限=" << rl << "  上限=" << ru << "\n";
}

void htestclass::put_utest(ostream& fout, option& opt, vecdbl& tvec, const ftable& ftab) const
// U検定の結果を出力する
{
  double u = 0.0, vu = 0.0, z = 0.0;
  testtype utype = utest(u, vu, z, ftab, opt.ut());
  ldbl pr = ztop(z);
  fout << "\n・Wilcoxonの2標本検定(順位和検定,Mann-WhitneyのU検定";
  if (utype == testcor) fout << ",修正有";
  fout << ")\n";
  fout << "  z=" << z << "  有意確率:p=" << pr << pind(pr) << "\n";
  double n1 = (double)(ftab.frtotal(1));
  double r1 = tvec[1] / n1;
  string code1(ftab.rcode(1));
  double n2 = (double)(ftab.frtotal(2));
  double r2 = tvec[2] / n2;
  string code2(ftab.rcode(2));
  fout << "  順位平均値:[" << code1 << "]=" << r1 << " [" << code2 << "]=" << r2 << "\n";
  double drm = r2 - r1;
  if (sgn(drm) == 0) drm = 0.0;
  double vd = vu / n1 / n2 * (n1 + n2) * (1.0 / n1 + 1.0 / n2);
  double ci = calmcci(mcfis, opt.cc(), ftab.nr(), 1, 0, 1, vd);
  double dl = drm - ci, du = drm + ci;
  fout << "  順位平均値の差の" << opt.cc() * 100.0 << "%信頼区間=" << drm << "±" << ci << "(" << dl << "-" << du << ")\n";
  double pro = 100.0 / n1 / n2, umax = n1 * n2;
  double u1 = umax - u;
  fout << "  U値(%):[" << code1 << "]=" << u1 << "(" << u1 * pro << "%)";
  fout << "  [" << code2 << "]=" << u << "(" << u * pro << "%)\n";
  ci = calmcci(mcfis, opt.cc(), ftab.nr(), 1, 0, 1, vu);
  dl = u - ci;  du = u + ci;
  if (dl < 0.0) dl = 0.0;
  if (du > n1 * n2) du = n1 * n2;
  fout << "  [" << code2 << "]のUの" << opt.cc() * 100.0 << "%信頼区間=" << u << "±" << ci << "(" << dl << "(" << dl * pro << "%)-" << du << "(" << du * pro << "%))\n";
}

void htestclass::put_bun(ostream& fout, option& opt, const ftable& ftab) const
// 2×2分割表に関する多重比較結果を出力する
{
  int dc = 1;
  string method = "χ2乗検定(連続修正有)";
  if (opt.bt() == mcbon)
  { method = "Bonferroni型多重比較(χ2乗検定:連続修正有)";  dc = ftab.nc() - 1; }
  fout << "\n○2×2分割表に対する" << method << "\n";
  ftable ftabj(ftab);           // 計算用ftable
  ftabj.nr(2);  ftabj.nc(2);    // 2x2分割表にする
  int n1 = ftab.frtotal(1), n2 = ftab.frtotal(2), df = 1;
  int fcum11 = 0, fcum12 = 0, fcum21 = 0, fcum22 = 0;
  double csqr = 0.0;
  for (size_t j = 1; j < ftab.nc(); ++j)
  {
    fcum11 += ftab.f(1, j), fcum21 += ftab.f(2, j);
    fcum12 = n1 - fcum11;  fcum22 = n2 - fcum21;
    ftabj.f(1, 1, fcum11);  ftabj.f(1, 2, fcum12);
    ftabj.f(2, 1, fcum21);  ftabj.f(2, 2, fcum22);
    ftabj.set();
    testtype ctype = ctest(csqr, df, ftabj, true);
    ldbl pr = ctop(csqr, df) * dc;
    if (pr > 1.0L) pr = 1.0L;
    string ccode(ftab.ccode(j + 1));
    strtrm(ccode);
    double p1 = (double)fcum12 / (double)n1, p2 = (double)fcum22 / (double)n2;
    fout << "・" << ccode << " 以上の出現率:" << ftab.rcode(2) << "群=" << p2;
    fout << " - " << ftab.rcode(1) << "群=" << p1 << "\n";
    fout << "  Χ2乗=" << csqr << "  自由度=" << df << "  有意確率:p=" << pr << pind(pr) << "\n";
    double d = 0.0;
    double ci = caldpci(d, opt.bt(), opt.cc(), ftab.nc(), 1, dc, fcum11, fcum12, fcum21, fcum22, true); // 連続補正あり
    fout << "  出現率の差の" << opt.cc() * 100.0 << "%信頼区間=" << d << "±" << ci << "(" << d - ci << "-" << d + ci << ")\n";
  }
}

void htestclass::put_hmcon(ostream& fout, mctype mc, bool ma, double kc, const vecdbl& tvec, const ftable& ftab) const
// 順位平均値の多重比較を出力する
{
  int df = 1, dc = 1;
  string method("Tukey");
  switch (mc)
  {
  case mcbon : method = "Bonferroni";
    dc = ma ? ftab.nr() * (ftab.nr() - 1) / 2 : ftab.nr() - 1;
    break;
  case mcdun : method = "Dunnett";
    break;
  case mcfis : method = "Fisher";
    break;
  case mcsch : method = "Scheffe";
    df = ftab.nr() - 1;
  }
  fout << "\n・群の" << method << "型多重比較(" << method << " type multiple comparison)\n";
  int llen = 72;                // 区切り線の長さ
  fout << "群       :順位平均 - 群       :順位平均";
  if (mc == mctuk) { fout << "       q値 群数";  llen += 5; }
  else if (mc == mcdun) { fout << "       d値 群数";  llen += 5; }
  else fout << "    χ2乗値";
  fout << " 自由度 有意確率p値\n";
  putline(fout, llen);
  size_t iend = ma ? ftab.nr() : 2;
  for (size_t i = 1; i < iend; ++i)
  {
    double ni = (double)(ftab.frtotal(i));
    double ri = tvec[i] / ni;
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      double nj = (double)(ftab.frtotal(j));
      double rj = tvec[j] / nj;
      double drm = rj - ri;
      double csqr = 6.0 * drm * drm / (1.0 / ni + 1.0 / nj) / tvec[0] / kc;
      double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
      ldbl pr = calmcpr(mc, csqr, ftab.nr(), df, 0, dc, r);
      int npad = ftab.rcode(j).size() - strclen(ftab.rcode(j), CHARwide);
      fout << left << setw(9 + npad) << ftab.rcode(j) << ":";
      fout << right << setw(8) << rj;
      fout << left << " - " << setw(9 + npadi) << codei << ":";
      fout << right << setw(8) << ri;
      if (mc == mctuk || mc == mcdun) fout << setw(10) << csqr << setw(5) << ftab.nr() << setw(7) << "inf";
      else fout << setw(10) << csqr << setw(7) << df;
      fout << setw(12) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, llen);
}

void htestclass::put_hmrdci(ostream& fout, option& opt, mctype mc, bool ma, double kc, const vecdbl& tvec, const ftable& ftab) const
// 2群間の順位平均値の差の信頼区間出力
{
  int df = 1, dc = 1;
  string method("Tukey");
  switch (mc)
  {
  case mcbon : method = "Bonferroni";
    dc = ma ? ftab.nr() * (ftab.nr() - 1) / 2 : ftab.nr() - 1;
    break;
  case mcdun : method = "Dunnett";
    break;
  case mcfis : method = "Fisher";
    break;
  case mcsch : method = "Scheffe";
    df = ftab.nr() - 1;
  }
  fout << "\n・" << method << "型" << opt.cc() * 100.0 << "%同時信頼区間(simultaneous confidence interval)\n";
  fout << "群       - 群       順位平均の差    絶対精度        下限        上限\n";
  const int LENline = 68;       // 区切り線の長さ
  putline(fout, LENline);
  size_t iend = ma ? ftab.nr() : 2;
  double deps = DBL_EPSILON * (double)(ftab.ftotal());
  for (size_t i = 1; i < iend; ++i)
  {
    double ni = (double)(ftab.frtotal(i));
    double ri = tvec[i] / ni;
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      double nj = (double)(ftab.frtotal(j));
      double rj = tvec[j] / nj;
      double drm = rj - ri;
      if (fabs(drm) < deps) drm = 0.0;
      double vd = kc * tvec[0] * (1.0 / ni + 1.0 / nj) / 6.0;
      double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
      double ci = calmcci(mc, opt.cc(), ftab.nr(), df, 0, dc, vd, r);
      int npad = (ftab.rcode(j)).size() - strclen(ftab.rcode(j), CHARwide);
      fout << left << setw(9 + npad) << ftab.rcode(j);
      fout << "- ";
      fout << setw(9 + npadi) << codei << right << setw(12) << drm;
      fout << setw(12) << ci << setw(12) << drm - ci << setw(12) << drm + ci;
      fout << "\n";
    }
  }
  putline(fout, LENline);
}

int htestclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("実行時エラー！");
  if (opt.check(nftab) != 0) throw option_error("実行時オプションエラー！");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void htestclass::put(option& opt, listidd& idlst, listftab& ftablst) const
// 結果を出力する
{
  if (opt.res("res") == false) return;
  listidd_itr iitr = idlst.begin();
  for (auto& ftabi : ftablst)
  {
    put_id(cout, *iitr);
    ftabi.trm();                // 行・列の合計が0のものを削除する
    if (ftabi.nr() > 0)
    {
      if (opt.res("tab")) putftable(cout, ftabi, opt.tab());
      if (ftabi.nr() > 1 && ftabi.nc() > 1)
      {
        double csqr = 0.0, kc = 0.0;
        vecdbl tvec(ftabi.nr() + 1);
        int df = htest(csqr, kc, tvec, ftabi);
        if (opt.res("r") || (opt.res("def") && ftabi.nr() > 2)) put_rr(cout, opt, df, csqr, ftabi); // 順位相関比出力
        if (opt.res("ht") || (opt.res("def") && ftabi.nr() > 2))
        {
          ldbl pr = ctop(csqr, df);
          cout << "\n・Kruskal-Wallisの検定\n";
          cout << "  H=χ2乗=" << csqr << "  自由度=" << df << "  有意確率:p=" << pr << pind(pr) << "\n";
        }
        if (ftabi.nr() == 2)
        {
          if (opt.res("ut") || opt.res("def")) put_utest(cout, opt, tvec, ftabi);
          if (opt.res("bt")) put_bun(cout, opt, ftabi);
        }
        else
        {
          if (opt.res("ma"))
          {
            put_hmcon(cout, opt.ma(), true, kc, tvec, ftabi);
            put_hmrdci(cout, opt, opt.ma(), true, kc, tvec, ftabi);
          }
          if (opt.res("md"))
          {
            put_hmcon(cout, opt.md(), false, kc, tvec, ftabi);
            put_hmrdci(cout, opt, opt.md(), false, kc, tvec, ftabi);
          }
        }
      }
      cout << "\n" << endl;
    }
    ++iitr;
  }
}

int main(int argc, char* argv[])
// メインルーチン
{
  htestclass htest;             // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          htest.get(nullptr, opt, idlst, ftablst);
          htest.put(opt, idlst, ftablst);
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        htest.get(argv[i], opt, idlst, ftablst);  fok = false;
        htest.put(opt, idlst, ftablst);
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
