// spear.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2024-12-16 10:28:01 tac>
//
// クロス集計表を入力し、Spearmanの順位相関係数を計算する。
// クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法  :g++ -O -s -W spear spear.C
//              -I/usr/local/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "spear  Version 7.42          Copyleft(Vo\\oV) 2024 杉本典夫",
  "書式:spear  [<オプション> …]  [<相関表ファイル> …]",
  "機能:相関表(クロス集計表)に対してSpearmanの順位相関係数を計算するよ。",
  "<オプション>",
  "  -cc[=<信頼係数>]:信頼係数を指定するよ。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込むよ。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するよ。(-h:-h=man)",
  "  -res=[-][{def|all|jt|k|r|st|tab[{fr|pc|pr|pt}]|w}][,…]",
  "    :結果の出力項目を指定するよ。",
  "    -:-を付けた項目を出力しないよ(-res=-:全ての項目を出力しないよ)",
  "    def:オプション省略時と同じ項目を出力するよ  all:全ての項目を出力するよ",
  "    jt:Jonckheere-Terpstra検定  k:一致係数κ  r:順位相関係数",
  "    st:交差度数の優劣比較",
  "    tab[{fr|pc|pr|pt}]:相関表",
  "      fr:度数だけ出力  pr:行合計に対する％も出力  pc:列合計に対する％も出力",
  "      pt:全合計に対する％も出力(オプション省略時)",
  "    w:Kendallの一致係数W",
  "    オプション省略時:データに合わせて自動判定"
};

const string Title = "=== 順序データの相関 ==="; // 表題

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  lstr vreslst;                 // 結果出力リスト
  mapsb vresmap;                // 結果出力の有無マップ
  tabtype vtab;                 // 度数集計表の種類
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  bool res(const string& code); // 解析結果出力フラッグを返す
  tabtype tab() const           // 度数集計表出力フラッグを返す
  { return vtab; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check();                  // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vreslst(), vresmap(), vtab(tabfprt)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  mapsb_citr resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check()             // オプションのチェックとセット
{
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["r"] = vresmap["tab"] = true;
  vresmap["jt"] = vresmap["k"] = vresmap["st"] = vresmap["w"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["r"] = vresmap["tab"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr resitr = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(resstr.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適当！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def") // 省略時出力
      { vresmap["def"] = vresmap["r"] = vresmap["tab"] = true; }
      else if (resstr == "all") // 全て出力
      {
        vresmap["def"] = vresmap["jt"] = vresmap["k"] = vresmap["r"] = vresmap["st"] = vresmap["tab"] = vresmap["w"] = true;
      }
      else if (resstr == "tabfr")  // 度数だけ出力
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (resstr == "tabpc")  // 列合計に対する％も出力
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (resstr == "tabpr")  // 行合計に対する％も出力
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (resstr == "tabpt")  // 全体合計に対する％も出力
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適当！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

class spearclass                // spearメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_rs(ostream& fout, option& opt, double rs, const ftable& ftab) const;
  // Spearmanの順位相関係数を出力する
  void put_w(ostream& fout, option& opt, double rs, const ftable& ftab) const;
  // Kendallの一致係数Wを出力する
  void put_jt(ostream& fout, const ftable& ftab) const;
  // Jonckheere-Terpstra検定結果を出力する
  void put_k(ostream& fout, option& opt, const ftable& ftab) const;
  // 一致係数κを出力する
  void put_st(ostream& fout, option& opt, const ftable& ftab) const;
  // 交差度数の優劣比較の結果を出力する
  void put_stpci(ostream& fout, option& opt, int f1, int f2) const;
  // 交差度数の出現率の信頼区間出力
public:
  spearclass() : fname()        // コンストラクタ
  {}
  ~spearclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // 度数集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst) const;
  // 結果を出力する
};

void spearclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "順序項目(縦):" << id.item(i) << "\n";
      break;
    case avalcol : fout << "順序項目(横):" << id.item(i) << "\n";
      break;
    default : id.put_item(fout, i);
    }
  }
}

void spearclass::put_rs(ostream& fout, option& opt, double rs, const ftable& ftab) const
// Spearmanの順位相関係数を出力する
{
  fout << "\n・Spearmanの順位相関係数";
  if (ftab.nr() == 2 && ftab.nc() == 2) fout << "(4分点相関係数)";
  fout << ":rs=" << rs << "  寄与率:rs2乗=" << rs * rs << "\n";
  double zr = 0.0;
  double ci = calrci(zr, opt.cc(), ftab.ftotal(), rs);
  fout << "  rsの" << opt.cc() * 100.0 << "%信頼区間:tanh(" << zr << "±" << ci << ")=" << tanh(zr - ci) << "-" << tanh(zr + ci) << "\n";
  double z = rs * sqrt((double)(ftab.ftotal() - 1));
  ldbl pr = ztop(z);
  fout << "  正規検定:z=" << z << "  有意確率:p=" << pr << pind(pr) << "\n";
  int df = ftab.ftotal() - 2;
  double t = (rs != 1.0) ? rs * sqrt((double)df / (1.0 - rs * rs)) : 0.0;
  pr = ttop(t, df);
  fout << "  t検定:t=" << t << "  自由度=" << df << "  有意確率:p=" << pr << pind(pr) << endl;
}

void spearclass::put_w(ostream& fout, option& opt, double rs, const ftable& ftab) const
// Kendallの一致係数Wを出力する
{
  double w = (rs + 1.0) * 0.5;
  double csqr = w * (double)(2 * ftab.ftotal() - 2);
  int df = ftab.ftotal() - 1;
  ldbl pr = ctop(csqr, df);
  fout << "\n・Kendallの一致係数:W=" << w << "\n";
  double zr = 0.0;
  double ci = calrci(zr, opt.cc(), ftab.ftotal(), rs);
  double wl = (tanh(zr - ci) + 1.0) * 0.5, wu = (tanh(zr + ci) + 1.0) * 0.5;
  fout << "  Wの" << opt.cc() * 100.0 << "%信頼区間:下限=" << wl << "  上限=" << wu << "\n";
  fout << "  Friedmanの検定:χ2乗=" << csqr << "  自由度=" << df << "  有意確率:p=" << pr << pind(pr) << endl;
}

void spearclass::put_jt(ostream& fout, const ftable& ftab) const
// Jonckheere-Terpstra検定結果を出力する
{
  double jt = 0.0, tmax = 0.0, vt = 0.0;
  double z = jttest(ftab, jt, tmax, vt);
  ldbl pr = ztop(z);
  double jtpr = (tmax > 0.0) ? 100.0 * jt / tmax : 0.0;
  fout << "\n・Jonckheere-Terpstra検定:JT(%)=" << jt << "(" << jtpr << "%) Tmax=" << tmax << "\n";
  fout << "  J=z=" << z << "  有意確率:p=" << pr << pind(pr) << endl;
}

void spearclass::put_k(ostream& fout, option& opt, const ftable& ftab) const
// 一致係数κを出力する
{
  double vk = 0.0, vk0 = 0.0, ci = 0.0;
  double k = kappa(vk, vk0, ftab, false);
  fout << "\n・一致係数:κ=" << k << "\n";
  ci = ptoz(1.0 - opt.cc()) * sqrt(vk);
  fout << "  κの" << opt.cc() * 100.0 << "%信頼区間:下限=" << k - ci << "  上限=" << k + ci << "\n";
  double z = (vk0 == 0.0) ? 0.0 : k / sqrt(vk0);
  ldbl pr = ztop(z);
  fout << "  正規近似検定:z=" << z << "  有意確率:p=" << pr << pind(pr) << "\n";
  k = kappa(vk, vk0, ftab, true);
  if (ftab.nr() == 2 && ftab.nc() == 2) return;
  fout << "\n・重み付き一致係数:κw=" << k << "\n";
  ci = ptoz(1.0 - opt.cc()) * sqrt(vk);
  fout << "  κwの=" << opt.cc() * 100.0 << "%信頼区間:下限=" << k - ci << "  上限=" << k + ci << "\n";
  z = (vk0 == 0.0) ? 0.0 : k / sqrt(vk0);
  pr = ztop(z);
  fout << "  正規近似検定:z=" << z << "  有意確率:p=" << pr << pind(pr) << endl;  
}

void spearclass::put_st(ostream& fout, option& opt, const ftable& ftab) const
// 交差度数の優劣比較の結果を出力する
{
  int f1 = ftab.f(1,2), f2 = ftab.f(2,1);
  int n = f1 + f2, dif = f1 - f2;
  double pr1 = (double)f1 / (double)n, pr2 = (double)f2 / (double)n;
  fout << "\n・交差度数の優劣比較=" << f1 << "(" << 100.0 * pr1 << "%):" << f2 << "(" << 100.0 * pr2 << "%)\n";
  if (0.0 < opt.cc() && opt.cc() < 1.0) put_stpci(fout, opt, f1, f2);
  ldbl pr0 = binom(f1, f2);
  ldbl pr = (pr0 < 0.5L) ? 2.0L * pr0 : 1.0L;
  fout << "  符号検定 有意確率:片側p=" << pr0 << pind(pr0) << " 両側p=" << pr << pind(pr) << "\n";
  double z = (double)(dif - sgn((double)dif)) / sqrt((double)n);
  pr = ztop(z);
  fout << "  McNemarの検定(修正有):z=" << z << "  両側有意確率:p=" << pr << pind(pr) << endl;
}

void spearclass::put_stpci(ostream& fout, option& opt, int f1, int f2) const
// 交差度数の出現率の信頼区間出力
{
  int n = f1 + f2, dif = f1 - f2;
  double p0 = (double)f1 / (double)n;
  double plow = 0.0, pup = 1.0;
  double pci = calbexci(plow, pup, f1, n, opt.cc());
  fout << "  交差度数" << f1 << "の出現率と理論確率の差=" << p0 << "-0.5=" << p0 - 0.5 << "\n";
  fout << "  差の" << opt.cc() * 100.0 << "%信頼区間:下限=" << plow - 0.5 << "  上限=" << pup - 0.5 << endl;
}

int spearclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("実行時エラー！");
  if (opt.check() != 0) throw option_error("実行時オプションエラー！");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void spearclass::put(option& opt, listidd& idlst, listftab& ftablst) const
// 結果を出力する
{
  if (opt.res("res") == false) return;
  listidd_itr iitr = idlst.begin();
  for (auto& ftabi : ftablst)
  {
    put_id(cout, *iitr);
    if (ftabi.nr() > 0)
    {
      if (opt.res("tab")) putftable(cout, ftabi, opt.tab());
      if (ftabi.nr() > 1 && ftabi.nc() > 1)
      {
        double rs = spear(ftabi);
        if (opt.res("r")) put_rs(cout, opt, rs, ftabi); // 順位相関係数出力
        if (opt.res("w")) put_w(cout, opt, rs, ftabi);  // Kendallの一致係数W出力
        if (opt.res("k")) put_k(cout, opt, ftabi);      // 一致係数κ出力
        if (opt.res("jt")) put_jt(cout, ftabi); // Jonckheere-Terpstra検定結果出力
        if (ftabi.nr() == 2 && ftabi.nc() == 2)
        {
          if (opt.res("st")) put_st(cout, opt, ftabi); // 交差度数の優劣比較出力
        }
        cout << "\n" << endl;
      }
    }
    ++iitr;
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  spearclass spear;             // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          spear.get(nullptr, opt, idlst, ftablst);
          spear.put(opt, idlst, ftablst);
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        spear.get(argv[i], opt, idlst, ftablst);  fok = false;
        spear.put(opt, idlst, ftablst);
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
