// trend.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2024-12-16 10:35:05 tac>
//
// クロス集計表を入力し、Cochran-Armitageの傾向分析とロジット分析を行う。
// クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法 :g++ -O -s -W trend trend.C
//             -I/usr/local/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "trend  Version 7.42          Copyleft(Vo\\oV) 2024 杉本典夫",
  "書式:trend  [<オプション> …]  [<クロス集計表ファイル> …]",
  "機能:クロス集計表に対して傾向分析をするよ。",
  "<オプション>",
  "  -cc[=<信頼係数>]:信頼係数を指定するよ。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込むよ。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示するよ。(-h:-h=man)",
  "  -ngp={log|pl|pro}[,…]:Ngraph用出力データを指定するよ。",
  "    log:ロジスティック曲線  pl:出現率直線  pro:出現率",
  "    オプション省略時:全て指定",
  "  -ngpname[=<出力ファイル基本名>]:Ngraph用データの出力ファイル基本名を指定。",
  "    オプション省略時:Ngraph用データ出力せず",
  "    -ngpname:-ngpname=trend_ngp<2桁の連番>.dat",
  "  -res=[-][{def|all|ca|log[{c|w}]|tab[{fr|pc|pr|pt}]}]",
  "      [,…]:解析結果の出力項目を指定するよ。",
  "    -:-を付けた項目を出力しない(res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力する  all:全ての項目を出力する",
  "    ca:Cochran-Armitageの傾向分析",
  "    log[{c|w}]:ロジット分析(c:繰り返し補完法  w:実用ロジット法)",
  "    tab[{fr|p|pc|pr|pt}]:クロス集計表",
  "      fr:度数だけ出力  pr:行合計に対する％も出力(オプション省略時)",
  "      pc:列合計に対する％も出力  pt:全合計に対する％も出力",
  "    オプション省略時:データに合わせて自動判定",
  "  -xrange=<最小値>-<最小値>:X軸の最小値と最大値を指定します。"
};

const string Title = "=== 出現率の傾向分析 ==="; // 表題
const string NGPNAMEdef = "trend_ngp"; // 省略時Ngraphファイル基本名

enum logtype                    // ロジット分析の種類
{
  lognon,                       // 実行せず
  logcomp,                      // 繰り返し補完法
  logwork,                      // 実用ロジット法
  logetc                        // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  logtype vlog;                 // ロジット分析の種類
  string vngpname;              // Ngraphファイル基本名
  string vngpnamen;             // Ngraphファイル名(番号付き)
  lstr vngplst;                 // Ngraph出力リスト
  mapsb vngpmap;                // Ngraph出力の有無マップ
  lstr vreslst;                 // 解析結果出力リスト
  mapsb vresmap;                // 解析結果出力の有無マップ
  tabtype vtab;                 // 度数集計表の種類
  pairdbl vxrange;              // X軸の最小値と最大値
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  logtype log() const           // ロジット分析の種類を返す
  { return vlog; }
  const string& ngpname(int fno); // Ngraphデータファイル名を返す
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 解析結果出力フラッグを返す
  tabtype tab() const           // 度数集計表の種類を返す
  { return vtab; }
  double xmin() const           // X軸の最小値を返す
  { return vxrange.first; }
  double xmax() const           // X軸の最大値を返す
  { return vxrange.second; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check();                  // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vlog(logcomp), vngpname(), vngpnamen(), vngplst(), vngpmap(), vreslst(), vresmap(), vtab(tabfprr), vxrange(BLANKf, BLANKf)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

const string& option::ngpname(int fno) // Ngraphデータファイル名を返す
{
  if (vngpname == "") return vngpname; // -ngpname指定なし
  ostringstream ostr;
  if (fno < 10) ostr << vngpname << "0" << fno << ".dat";
  else ostr << vngpname << fno << ".dat";
  vngpnamen = (string)(ostr.str());
  return vngpnamen;             // 番号付きファイル名を返す
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  mapsb_citr ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  mapsb_citr resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル基本名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "xrange")  // X軸の最小値と最大値の指定
  {
    if (getrange(vxrange.first, vxrange.second, opt[1]) == 0) // 範囲指定無
    { err = badopt(vpname, optstr, USAGE, -DBL_MAX, DBL_MAX); }
    else if (vxrange.first >= vxrange.second) // 範囲指定不適切
    { err = badopt(vpname, optstr, USAGE, -DBL_MAX, DBL_MAX); }
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check()             // オプションのチェックとセット
{
  if (set_ngp() < 0) return -2; // -ngpオプションのエラー
  if (set_res() < 0) return -3; // -resオプションのエラー
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["log"] = vngpmap["pl"] = vngpmap["pro"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "-ngpnameオプションの指定が必要！");
    vngpmap["log"] = vngpmap["pl"] = vngpmap["pro"] = false;
    for (const string& ngpstr : vngplst)
    {
      mapsb_citr ngpitr = vngpmap.find(ngpstr);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "-ngp=" + ngpstr + "オプション指定が不適当！"); }
      else
      { vngpmap[ngpstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["ca"] = vresmap["tab"] = true;
  vresmap["log"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["ca"] = vresmap["tab"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr resitr = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(resstr.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適当！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def") // 省略時出力
      { vresmap["def"] = vresmap["ca"] = vresmap["tab"] = true; }
      else if (resstr == "all") // 全て出力
      {
        vresmap["def"] = vresmap["ca"] = vresmap["tab"] = vresmap["log"] = true;
        vlog = logcomp;  vtab = tabfprr;
      }
      else if (resstr == "logc")
      { vresmap["log"] = true;  vlog = logcomp; }
      else if (resstr == "logw")
      { vresmap["log"] = true;  vlog = logwork; }
      else if (resstr == "tabfr")
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (resstr == "tabpr")
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (resstr == "tabpc")
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (resstr == "tabpt")
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適当！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

class trendclass                // trendメインクラス
{
private:
  string fname;                 // ファイル名
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  void put_id(ostream& fout, idd& id) const;
  // 度数表のタイトルとデータ識別情報を出力する
  void put_trend(ostream& fout, option& opt, const ftable& ftab) const;
  // 回帰係数の信頼区間を出力する
  void put_ccoef(ostream& fout, option& opt, double b, double sxxw) const;
  // 回帰係数の信頼区間を出力する
  void put_cor(ostream& fout, option& opt, double b, double sxxw) const;
  void put_catab(ostream& fout, const string& head, double syyw, double sb, int dft) const;
  // Cochran-Armitageの傾向分析の分散分析表を出力する
  int cal_caline(double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab) const;
  // Cochran-Armitageの傾向分析
  int cal_lpline(option& opt, double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab) const;
  // ロジット分析の繰り返し計算
  int cal_logit(double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab, bool comp) const;
  // ロジット分析
  void plot_ngp(ofstream& fout, option& opt, idd& id, const ftable& ftab) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngplog(ofstream& fout, option& opt, double a, double b, const ftable& ftab) const;
  // Ngraph用にロジスティック曲線を出力する
  void plot_ngppl(ofstream& fout, option& opt, double a, double b, const ftable& ftab) const;
  // Ngraph用に出現率直線を出力する
  void plot_ngppro(ofstream& fout, const ftable& ftab) const;
  // Ngraph用に出現率を出力する
public:
  trendclass() : fname(), idlst(), ftablst() // コンストラクタ
  {}
  ~trendclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // 度数集計表を読み込む
  void put(option& opt, int fno);            // 結果を出力する
};

void trendclass::put_id(ostream& fout, idd& id) const
// 度数表のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "説明変数項目(縦):" << id.item(i) << "\n";
      break;
    case avalcol : fout << "目的変数項目(横):" << id.item(i) << "\n";
      break;
    default : id.put_item(fout, i);
    }
  }
}

void trendclass::put_trend(ostream& fout, option& opt, const ftable& ftab) const
// 回帰係数の信頼区間を出力する
{
  string ccode = ftab.ccode(2);
  strtrm(ccode);
  double a = 0.0, b = 0.0, sxxw = 0.0, syyw = 0.0, sb = 0.0;
  int nr = 0, dft = 0;
  bool puttab = false;
  if (opt.res("ca"))            // Cochran-Armitageの傾向分析出力
  {
    nr = cal_caline(a, b, sxxw, syyw, sb, dft, ftab);
    fout << "・横のコード=" << ccode << "の出現率直線:p=" << a << showpos << b << noshowpos << "x\n";
    put_ccoef(fout, opt, b, sxxw);
    double r2 = sb / syyw;
    fout << "  寄与率r2乗=" << r2 << "  r=" << sqrt(r2) << "\n";
    put_catab(fout, "Cochran-Armitageの傾向分析", syyw, sb, dft);
    puttab = true;
  }
  if (opt.res("log"))           // ロジット分析出力
  {
    nr = cal_lpline(opt, a, b, sxxw, syyw, sb, dft, ftab);
    if (puttab) fout << "\n";
    fout << "・横のコード=" << ccode << "の出現率曲線:p=1/{1+exp(" << -a << showpos << -b << noshowpos << "x)}\n";
    fout << "  p0=1/{1+exp(" << -a << ")}=" << 1.0 / (1.0 + exp(-a));
    if (nr > 0) fout << "  (繰り返し補完回数=" << nr << ")";
    fout << "\n";
    put_ccoef(fout, opt, b, sxxw);
    put_cor(fout, opt, b, sxxw);
    double r2 = sb / syyw;
    fout << "  寄与率r2乗=" << r2 << "  r=" << sqrt(r2) << "\n";
    put_catab(fout, "ロジット分析", syyw, sb, dft);
  }
}

void trendclass::put_ccoef(ostream& fout, option& opt, double b, double sxxw) const
// 回帰係数の信頼区間を出力する
{
  double z = ptoz(1.0 - opt.cc());
  double ccpr = opt.cc() * 100.0;
  double ci = (sxxw > 0.0) ? z * sqrt(1.0 / sxxw) : 0.0;
  fout << "  回帰係数の" << ccpr << "%信頼区間=";
  fout << b << "±" << ci << "(" << b - ci << "-" << b + ci << ")\n";
}

void trendclass::put_cor(ostream& fout, option& opt, double b, double sxxw) const
// オッズ比の信頼区間を出力する
{
  double z = ptoz(1.0 - opt.cc());
  double ccpr = opt.cc() * 100.0;
  double ci = (sxxw > 0.0) ? z * sqrt(1.0 / sxxw) : 0.0;
  fout << "  オッズ比:OR=" << exp(b) << "  " << ccpr << "%信頼区間=";
  fout << exp(b - ci) << "-" << exp(b + ci) << endl;
}

void trendclass::put_catab(ostream& fout, const string& head, double syyw, double sb, int dft) const
// Cochran-Armitage分析またはロッジト分析の分散分析表を出力する
{
  int w = strclen(head, CHARwide);
  int npad = head.size() - w;
  fout << "\n" << setw((68 + w) / 2 + npad) << head;
  fout << "\n要因                平方和    自由度        χ2乗値   有意確率p値\n";
  constexpr int LENline = 68;
  putline(fout, LENline);
  int dfb = 1;
  int dfl = dft - dfb;
  double sl = syyw - sb;
  double eps = DBL_EPSILON * fmax(5.0, syyw);
  if (dfl < 1 || sl < eps) sl = 0.0;
  ldbl pr = ctop(sb, dfb);
  if (head == "Cochran-Armitageの傾向分析") fout << "直線性" << setw(20) << sb;
  else fout << "回帰" << setw(22) << sb;
  fout << setw(10) << dfb;
  fout << setw(14) << sb << setw(14) << pr << pind(pr) << "\n";
  pr = ctop(sl, dfl);
  fout << "異質性(ズレ)" << setw(14) << sl << setw(10) << dfl;
  fout << setw(14) << sl << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  pr = ctop(syyw, dft);
  fout << "全体" << setw(22) << syyw << setw(10) << dft;
  if (head == "Cochran-Armitageの傾向分析") fout << setw(14) << syyw << setw(14) << pr << pind(pr);
  fout << endl;
}

int trendclass::cal_caline(double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab) const
// Cochran-Armitageの傾向分析
{
  int sn = 0, sr = 0;
  double mx = 0.0, sxyw = 0.0;
  sxxw = syyw = sb = 0.0;  dft = 0;
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double x = strtodf(ftab.rcode(i));
    int n = ftab.frtotal(i), r = ftab.f(i, 2);
    if (x != BLANKf && n > 0)
    {
      ++dft;  sn += n;  sr += r;
      mx += (double)n * x;  sxxw += (double)n * x * x;
      syyw += (double)r * (double)r / (double)n;  sxyw += double(r) * x;
    }
  }
  --dft;
  if (dft < 1)
  {
    sxxw = syyw = sb = 0.0;
    return -1;                  // ロジット分析不可能
  }
  const double Seps = DBL_EPSILON * fmax(5.0, syyw);
  double pc = (double)sr / (double)sn;
  double w = 1.0 / pc / (1.0 - pc);
  mx /= (double)sn;
  sxxw = w * (sxxw - (double)sn * mx * mx);
  if (sxxw < Seps) sxxw = 0.0;
  syyw = w * (syyw - (double)sn * pc * pc);
  if (syyw < Seps) syyw = 0.0;
  sxyw = w * (sxyw - (double)sn * mx * pc);
  if (fabs(sxyw) < Seps) sxyw = 0.0;
  sb = sxyw * sxyw / sxxw;
  if (fabs(sb) < Seps) sb = 0.0;
  b = (sxxw > 0.0) ? sxyw / sxxw : 0.0;
  a = pc - b * mx;
  return dft;                   // 自由度を返す
}

int trendclass::cal_lpline(option& opt, double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab) const
// ロジット分析の繰り返し計算
{
 constexpr double EPSdef = 0.00001; // 省略時収束誤差
 constexpr int Rmax = 100;          // 繰り返し数最大値
 int nr = 0;
 int nest = cal_logit(a, b, sxxw, syyw, sb, dft, ftab, false);
  const double Eps = syyw * EPSdef;
  if (nest > 0 && opt.log() == logcomp) // 繰り返し補完法
  {
    double syywpre = syyw;
    nest = 0;
    bool comp = true;
    while (comp)
    {
      nest = cal_logit(a, b, sxxw, syyw, sb, dft, ftab, true);
      ++nr;
      if (nest < 0 || fabs(syyw - syywpre) < Eps || nr >= Rmax) comp = false;
      else syywpre = syyw;
    }
    double at = a, bt = b;    // 収束後のaとbで平方和を求める
    nest = cal_logit(at, bt, sxxw, syyw, sb, dft, ftab, true);
  }
  if (nest < 0) return -1;      // 計算できず
  return nr;                    // 繰り返し数を返す
}

int trendclass::cal_logit(double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab, bool comp) const
// ロジット分析
{
  double snw = 0.0, mxw = 0.0, myw = 0.0, sxyw = 0.0;
  sxxw = syyw = sb = 0.0;  dft = 0;
  int nest = 0;                 // 推測値で補完した用量数
  double y = 0.0, nw = 0.0;
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double x = strtodf(ftab.rcode(i));
    int n = ftab.frtotal(i), r = ftab.f(i, 2);
    if (x != BLANKf && n > 0)
    {
      if (0 < r && r < n)
      {
        y = log((double)r / (double)(n - r));
        nw = (double)r / (double)n * (double)(n - r);
      }
      else
      {
        if (comp)               // 推測値で補完する
        {
          double p = 1.0 / (1.0 + exp(-a - b * x));
          y = log(p / (1.0 - p));  nw = (double)n * p * (1.0 - p);
        }
        else                    // 実用ロジットを用いる
        {
          y = log(((double)r + 0.5) / ((double)(n - r) + 0.5));
          nw = ((double)r + 0.5) / (double)(n + 1) * ((double)(n - r) + 0.5);
        }
        ++nest;
      }
      ++dft;  snw += nw;  mxw += nw * x;  sxxw += nw * x * x;
      myw += nw * y;  syyw += nw * y * y;  sxyw += nw * x * y;
    }
  }
  --dft;
  const double Seps = DBL_EPSILON * fmax(5.0, syyw);
  if (dft < 1 || snw < Seps)
  {
    sxxw = syyw = sb = 0.0;
    return -1;                  // ロジット分析不可能
  }
  sxxw = sxxw - mxw * mxw / snw;
  if (sxxw < Seps) sxxw = 0.0;
  syyw = syyw - myw * myw / snw;
  if (syyw < Seps) syyw = 0.0;
  sxyw = sxyw - mxw * myw / snw;
  if (fabs(sxyw) < Seps) sxyw = 0.0;
  sb = sxyw * sxyw / sxxw;
  if (fabs(sb) < Seps) sb = 0.0;
  mxw = mxw / snw;  myw = myw / snw;
  b = sxyw / sxxw;
  a = myw - b * mxw;
  return nest;                  // 推測値で補完した用量数を返す
}

void trendclass::plot_ngp(ofstream& fout, option& opt, idd& id, const ftable& ftab) const
// Ngraph用データを出力する
{
  double a = 0.0, b = 0.0, sxxw = 0.0, syyw = 0.0, sb = 0.0;
  int dft = 0, nr = 0;
  bool delml = false;
  if (opt.ngp("pro"))           // 出現率
  {
    fout << "-s1 -dmark,00 -o300 -zy0,1,0.2\n";
    plot_ngpid(fout, id);
    plot_ngppro(fout, ftab);
    delml = true;
  }
  if (opt.ngp("log"))           // ロジスティック曲線
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    nr = cal_lpline(opt, a, b, sxxw, syyw, sb, dft, ftab);
    plot_ngplog(fout, opt, a, b, ftab);
  }
  if (opt.ngp("pl"))            // 出現率直線
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -w80 -zy0,1,0.2\n";
    plot_ngpid(fout, id);
    nr = cal_caline(a, b, sxxw, syyw, sb, dft, ftab);
    plot_ngppl(fout, opt, a, b, ftab);
  }
}

void trendclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "%説明変数項目(縦):" << id.item(i) << "\n";
      break;
    case avalcol : fout << "%目的変数項目(横):" << id.item(i) << "\n";
      break;
    default : id.put_item(fout, i, formngp);
    }
  }
}

void trendclass::plot_ngplog(ofstream& fout, option& opt, double a, double b, const ftable& ftab) const
// Ngraph用にロジスティック曲線を出力する
{
  fout << "%ロジスティック曲線\n";
  constexpr uint RDdef = 50;    // ロジスティック曲線の分解数
  double xmg = 0.1;             // 描画範囲のマージン
  double xmin = strtodf(ftab.rcode(1)), xmax = strtodf(ftab.rcode(ftab.nr()));
  if (opt.xmin() != BLANKf && opt.xmax() != BLANKf)
  { xmin = opt.xmin();  xmax = opt.xmax();  xmg = 0.0; }
  double range = xmax - xmin;
  double xstart = xmin - range * xmg, xend = xmax + range * xmg;
  double dx = (xend - xstart) / (double)RDdef;
  double x = xstart;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double pr = 1.0 / (1.0 + exp(-a - b * x));
    fout << x << " " << pr << "\n";
  }
}

void trendclass::plot_ngppl(ofstream& fout, option& opt, double a, double b, const ftable& ftab) const
// Ngraph用に出現率直線を出力する
{
  fout << "%出現率直線\n";
  double xmin = strtodf(ftab.rcode(1)), xmax = strtodf(ftab.rcode(ftab.nr()));
  double xmg = 0.1;             // 描画範囲のマージン
  if (opt.xmin() != BLANKf && opt.xmax() != BLANKf)
  { xmin = opt.xmin();  xmax = opt.xmax();  xmg = 0.0; }
  double range = xmax - xmin;
  double x = xmin - range * xmg;
  fout << x << " " << a + b * x << "\n";
  x = xmax + range * xmg;
  fout << x << " " << a + b * x << endl;
}

void trendclass::plot_ngppro(ofstream& fout, const ftable& ftab) const
// Ngraph用に出現率を出力する
{
  fout << "%出現率\n";
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double x = strtodf(ftab.rcode(i));
    double pro = (double)(ftab.f(i, 2)) / (double)(ftab.frtotal(i));
    fout << x << " " << pro << "\n";
  }
}

int trendclass::get(const char* c_fname, option& opt) // 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("実行時エラー！");
  if (opt.check() != 0) throw option_error("実行時オプションエラー！");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void trendclass::put(option& opt, int fno) // 結果を出力する
{
  listidd_itr iitr = idlst.begin();
  bool puttab = false;
  for (auto& ftabi : ftablst)
  {
    ftabi.trm();                // 行・列の合計が0のものを削除する
    if (ftabi.nr() > 0)
    {
      if (opt.res("res"))
      {
        put_id(cout, *iitr);
        if (opt.res("tab")) putftable(cout, ftabi, opt.tab());
        puttab = true;
      }
      if (ftabi.nr() > 1 && ftabi.nc() == 2)
      {
        double a = 0.0, b = 0.0, sxxw = 0.0, syyw = 0.0, sb = 0.0;
        int dft = 0, nr = 0;
        if (opt.res("res"))
        {
          if (opt.res("ca") || opt.res("log"))
          {
            if (puttab) cout << "\n";
            put_trend(cout, opt, ftabi);
          }
          cout << "\n" << endl;
        }
        if (opt.ngpname(fno) != "") // Ngraph用データ出力
        {
          ofstream foutngp;
          if (ofopen(foutngp, (opt.ngpname(fno)).c_str()) == false) return;
          plot_ngp(foutngp, opt, *iitr, ftabi);
          foutngp.close();
        }
      }
    }
    ++iitr;  ++fno;
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  trendclass trend;             // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int fno = 0;                // ファイル番号
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          int nftab = trend.get(nullptr, opt);
          trend.put(opt, fno);
          fno += nftab;
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        int nftab = trend.get(argv[i], opt);  fok = false;
        trend.put(opt, fno);
        fno += nftab;
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
