// wtab.C Version 7.42   Copyleft (c) 杉本典夫
// Time-stamp: <2024-07-22 15:59:59 tac>
//
// 順位データの差を集計し、符号付き順位検定用度数データを出力する。
// 度数データはwtestコマンドによって解析する。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法 :g++ -O -s -W wtab wtab.C -I/usr/local/include
//              -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "wtab  Version 7.42           Copyleft(Vo\\oV) 2024 杉本典夫",
  "書式:wtab  [<オプション> …]  [<データファイル> …]",
  "機能:Wilcoxonの符号付順位検定と符号検定用にグレードの差を集計するぜ。",
  "<オプション>",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込むぜ。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示ぜ。(-h:-h=man)",
  "  -key=<基準項目番号>:基準項目の番号を指定するぜ。(オプション省略時:-key=1)",
  "   -key=0:基準項目なし",
  "  -w=<出力桁数>:データの出力桁数を指定するぜ。(最小値:2)",
  "    オプション省略時:-w=6"
};

constexpr int Wdef = 6;         // 省略時出力桁数
constexpr int Wmax = LENcrt / 2; // 最大出力桁数

using mapdp = map<double, pairint>; // グレードマップの別名定義
using mapdp_itr = map<double, pairint>::iterator; // 指定子の別名定義
using mapdp_citr = map<double, pairint>::const_iterator; // const指定子の別名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  int vkey;                     // 基準項目番号
  int vw;                       // データの出力桁数
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  int key() const               // 基準項目番号を返す
  { return vkey; }
  int w() const                 // データの出力桁数を返す
  { return vw; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
};

option::option(const char* c_pname) : vpath(), vpname(), vw(Wdef), vkey(1)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "key")     // 基準項目番号指定
  {
    char* ptr = nullptr;
    vkey = strtol(opt[1].c_str(), &ptr, 0);
    if (vkey < 0) err = badopt(vpname, optstr, USAGE, 0.0, (double)INT_MAX);
  }
  else if (opt[0] == "w")       // データの出力桁数指定
  {
    char* ptr = 0;
    vw = strtol(opt[1].c_str(), &ptr, 0);
    if (vw < 2 || Wmax < vw) err = badopt(vpname, optstr, USAGE, 2, Wmax);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (vkey > (int)ni) return disperr(vpname, "基準項目番号が項目数より大きいぜ！");
  return 0;                     // エラーなし
}

class wtabclass                 // wtabメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int sum_wtab(vector<mapdp>& wmapvec, const option& opt);
  // グレードの度数集計
  void put_id(ostream& fout, const option& opt, int idx);
  // タイトルとデータ識別情報を出力する
public:
  wtabclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~wtabclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  void put(option& opt);        // 集計結果を出力する
};

int wtabclass::sum_wtab(vector<mapdp>& wmapvec, const option& opt)
// グレードの度数集計
{
  int ngrade = 0;
  for (auto& cdatai : datalst)
  {
    double gkey = (opt.key() > 0) ? cdatai.fdata(opt.key() - 1) : 0.0;
    if (gkey != BLANKf)
    {
      for (size_t j = 0; j < cdatai.n(); ++j)
      {
        int ino = j + 1;
        if (ino != opt.key())
        {
          double g = cdatai.fdata(j);
          double dg = g - gkey;
          if (g != BLANKf && sgn(dg) != 0)
          {
            double abs_dg = fabs(dg); // グレードの差/実数の絶対値
            mapdp_itr witr = wmapvec[j].find(abs_dg);
            if (witr == wmapvec[j].end()) // 新規データ
            {
              pairint tf(0, 0);
              if (dg > 0.0) tf.first = 1; // 差/実数が正の度数
              else tf.second = 1;     // 差/実数が負の度数
              wmapvec[j].insert(mapdp::value_type(abs_dg, tf));
            }
            else                      // 既存データ
            {
              if (dg > 0.0) ++(witr->second.first);
              else ++(witr->second.second);
            }
            ++ngrade;
          }
        }
      }
    }
  }
  return ngrade;                // 集計したグレード数を返す
}

void wtabclass::put_id(ostream& fout, const option& opt, int idx)
// タイトルとデータ識別情報を出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  if (opt.key() > 0) id.put_item(fout, opt.key() - 1, formtag);
  id.put_item(fout, idx, formtag);
}

int wtabclass::get(const char* c_fname, option& opt) // データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("実行時エラー！");
  if (opt.check(id.ni()) != 0) throw option_error("実行時オプションエラー！");
  return nc;                    // 読み込んだケース数を返す
}

void wtabclass::put(option& opt) // 集計結果を出力する
{
  vector<mapdp> wmapvec(id.ni()); // グレードマップベクトル
  if (id.nc() < 1 || sum_wtab(wmapvec, opt) < 1) return; // データなし
  datac tdata;
  ostringstream ostr;
  for (size_t j = 0; j < id.ni(); ++j)
  {
    int ino = j + 1;
    if (ino != opt.key())
    {
      if (wmapvec[j].size() > 0)
      {
        put_id(cout, opt, j);
        tdata.resize(wmapvec[j].size() + 1);
        if (opt.key() > 0) tdata.assign(0, "差の絶対値 ");
        else tdata.assign(0, "絶対値 ");
        int idx = 1;
        for (mapdp_citr witr = wmapvec[j].begin(); witr != wmapvec[j].end(); ++witr, ++idx)
        {
          ostr.str("");
          ostr << setw(opt.w()) << witr->first;
          tdata.assign(idx, ostr.str());
        }
        tdata.put(cout);
        if (opt.key() > 0) tdata.assign(0, "差:正の群 ");
        else tdata.assign(0, "正の群 ");
        idx = 1;
        for (mapdp_citr witr = wmapvec[j].begin(); witr != wmapvec[j].end(); ++witr, ++idx)
        {
          ostr.str("");
          ostr << witr->second.first;
          tdata.assign(idx, ostr.str());
        }
        tdata.put(cout);
        if (opt.key() > 0) tdata.assign(0, "差:負の群 ");
        else tdata.assign(0, "負の群 ");
        idx = 1;
        for (mapdp_citr witr = wmapvec[j].begin(); witr != wmapvec[j].end(); ++witr, ++idx)
        {
          ostr.str("");
          ostr << witr->second.second;
          tdata.assign(idx, ostr.str());
        }
        tdata.put(cout);
      }
      cout << endl;
    }
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  wtabclass wtab;               // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) wtab.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { wtab.get(argv[i], opt);  fok = false; }
    }
    wtab.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
