// wtest.C Version 7.42  Copyleft (c) 杉本典夫
// Time-stamp: <2024-12-16 10:43:00 tac>
//
// グレード差の集計表を入力し、Wilcoxonの1標本検定と符号検定を行う。
// 通常はwtabコマンドとパイプで接続して用いる。
// 3時期でも計算するが、その場合は本来はFriedmanの検定を用いるべきである。
//
// 使用ライブラリ:libdans.so, libfun.so
// コンパイル法 :g++ -O -s -W wtest wtest.C
//             -I/usr/local/include -L/usr/local/lib -ldans -lfun

#include <dans.H>               // dans用ヘッダファイル(fun.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "wtest  Version 7.42          Copyleft(Vo\\oV) 2024 杉本典夫",
  "書式:wtest [<オプション> …] [<グレードの差の集計表ファイル> …]",
  "機能:グレードの差/実測値の集計表に対して、Wilcoxonの1標本検定と",
  "     符号検定をしたるでな。",
  "<オプション>",
  "  -cc[=<信頼係数>]:信頼係数を指定したる。(0≦信頼係数<1)",
  "    -cc:-cc=0.95(オプション省略時)",
  "  -f=<ファイル名>:指定されたファイルからオプションを読み込んだる。",
  "  -h[=<コマンド名>]:オンラインマニュアルを表示したる。(-h:-h=man)",
  "  -res=[-][{def|all|m[<検定数>]|mt[{c|o}]|r|st",
  "    |tab[{fr|p|pc|pr|pt}]|wt[{a|c|o}]}][,…]:解析結果の出力項目を指定したる。",
  "    -:-を付けた項目を出力しない(-res=-:全ての項目を出力しない)",
  "    def:オプション省略時と同じ項目を出力したる  all:全ての項目を出力したる",
  "    m[<検定数>]:多重検定(m:自動設定 1＜検定数)",
  "    mt[{c|o}]:McNemar検定(c:連続修正有 o:連続修正無)",
  "    r:寄与率  st:符号検定",
  "    tab[={fr|p|pc|pr|pt}]:グレード差の集計表",
  "      fr:度数だけ  p:列または行合計に対する％も出力(オプション省略時)",
  "      pc:列合計に対する％も出力  pr:行合計に対する％も出力",
  "      pt:全合計に対する％も出力",
  "    wt[{a|c|o}]:Wilcoxonの1標本検定(符号付き順位検定)",
  "      a:連続修正自動判定  c:連続修正有  o:連続修正無",
  "    オプション省略時:データに合わせて自動判定"
};

const string Title = "=== 順序データの群内比較 ==="; // 表題

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  int vm0;                      // 多重検定時の指定検定数
  int vm;                       // 多重検定時の検定数
  testtype vmt;                 // McNemar検定結果の種類
  lstr vreslst;                 // 結果出力リスト
  mapsb vresmap;                // 結果出力の有無マップ
  tabtype vtab;                 // 度数集計表の種類
  testtype vwt;                 // Wilcoxonの1標本検定結果の種類
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  int m0() const                // 多重検定時の指定検定数を返す
  { return vm0; }
  int m() const                 // 多重検定時の検定数を返す
  { return vm; }
  int m(int m0)                 // 多重検定時の検定数を設定する
  { return (vm = m0); }
  testtype mt() const           // McNemar検定結果の種類を返す
  { return vmt; }
  bool res(const string& code); // 解析結果出力フラッグを返す
  tabtype tab() const           // 度数集計表出力フラッグを返す
  { return vtab; }
  testtype wt() const           // Wilcoxonの1標本検定結果の種類を返す
  { return vwt; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int nfile);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vm0(0), vm(0), vmt(testcor), vtab(tabetc), vwt(testaut)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 結果出力フラッグを返す
{
  mapsb_citr respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("ヘルプ表示");
  if (err < 0) throw option_error(string("オプション指定違反→") + optstr);
}

int option::check(int nftab)    // オプションのセットとチェック
{
  if (set_res() < 0) return -2; // -resオプションのセット
  if (vm0 == 0) vm = nftab;     // 多重検定数自動設定
  else vm = vm0;                // 多重検定数を指定値に設定する
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["tab"] = vresmap["wt"] = true;
  vresmap["m"] = vresmap["mt"] = vresmap["r"] = vresmap["st"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["tab"] = vresmap["wt"] = false;
    for (const string& resstr : vreslst)
    {
      mapsb_citr respos = vresmap.find(resstr);
      if (resstr == "-")        // 全て出力せず
      { vresmap["res"] = false; }
      else if (resstr[0] == '-' && resstr.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(resstr.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！"); }
        else
        { vresmap[resstr.substr(1)] = false; }
      }
      else if (resstr == "def") // 省略時出力
      { vresmap["def"] = vresmap["tab"] = vresmap["wt"] = true; }
      else if (resstr == "all") // 全て出力
      {
        vresmap["def"] = vresmap["m"] = vresmap["mt"] = vresmap["r"] = vresmap["st"] = vresmap["tab"] = vresmap["wt"] = true;
        vmt = testcor;  vtab = tabetc;  vwt = testaut;
      }
      else if (resstr[0] == 'm' && resstr.size() > 1)
      {
        char* ptr = nullptr;
        vm0 = strtol(resstr.substr(1).c_str(), &ptr, 0); // 検定数
        if (vm0 < 2) return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！");
        vresmap["m"] = true;
      }
      else if (resstr == "mtc") // 連続修正有り
      { vresmap["mt"] = true;  vmt = testcor; }
      else if (resstr == "mto") // 連続修正無し
      { vresmap["mt"] = true;  vmt= testorg; }
      else if (resstr == "tabfr") // 度数だけ出力
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (resstr == "tabp")  // 列または行合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabetc; }
      else if (resstr == "tabpc") // 列合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (resstr == "tabpr") // 行合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (resstr == "tabpt") // 全合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (resstr == "wta") // 連続修正自動判定
      { vresmap["wt"] = true;  vwt= testaut; }
      else if (resstr == "wtc") // 連続修正有り
      { vresmap["wt"] = true;  vwt= testcor; }
      else if (resstr == "wto") // 連続修正無し
      { vresmap["wt"] = true;  vwt = testorg; }
      else if (respos == vresmap.end())
      { return disperr(vpname, "-res=" + resstr + "オプション指定が不適切です！"); }
      else
      { vresmap[resstr] = true; }
    }
  }
  return 0;                     // エラーなし
}

class wtestclass                // wtestメインクラス
{
private:
  string fname;                 // ファイル名
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_wtest(ostream& fout, option& opt, const ftable& ftab) const;
  // Wilcoxonの1標本検定の結果を出力する
  void put_tpci(ostream& fout, option& opt, double tp, double vt, mctype mc, int dc, const ftable& ftab) const;
  // Wilcoxonの1標本検定のTpの信頼区間出力
  void put_stest(ostream& fout, option& opt, const ftable& ftab) const;
  // 符号検定の結果を出力する
  void put_stpci(ostream& fout, option& opt, mctype mc, int dc, const ftable& ftab) const;
  // 符号検定のpp信頼区間出力
  void put_mtest(ostream& fout, option& opt, const ftable& ftab) const;
  // Mcnemar検定の結果を出力する
  void put_mtpci(ostream& fout, option& opt, mctype mc, int dc, const ftable& ftab) const;
  // Mcnemar検定のpp信頼区間出力
public:
  wtestclass() : fname(), idlst(), ftablst() // コンストラクタ
  {}
  ~wtestclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // 度数集計表を読み込む
  void put(option& opt);                     // 結果を出力する
};

void wtestclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  if (id.ni() > 1)
  {
    fout << "集計項目 :" << id.item(1) << "\n";
    fout << "-比較項目:" << id.item(0) << "\n";
  }
  else
  { fout << "集計項目 :" << id.item(0) << "\n"; }
}

void wtestclass::put_wtest(ostream& fout, option& opt, const ftable& ftab) const
// Wilcoxonの1標本検定の結果を出力する
{
  double tp = 0.0, vt = 0.0;
  double z = wtest(ftab, tp, vt, opt.wt());
  ldbl pr = ztop(z);
  fout << "\n・Wilcoxonの1標本検定(符号付き順位検定";
  if (opt.wt() == testcor || (opt.wt() == testaut && fabs(z) < 2.0)) fout << ",修正有";
  fout << ")\n";
  fout << "  正規分布:z=" << z << "  有意確率:p=" << pr << pind(pr) << "\n";
  if (opt.res("r"))             // 寄与率出力
  {
    z = wtest(ftab, tp, vt, testorg);
    fout << "  寄与率:r2乗=" << z * z / double(ftab.ftotal()) << "\n";
  }
  put_tpci(fout, opt, tp, vt, mcfis, 1, ftab);
  if (opt.res("m") == false || opt.m() == 1) return;
  pr = pr * (ldbl)(opt.m());
  if (pr > 1.0L) pr = 1.0L;
  fout << "\n  Bonferroni型多重検定  有意確率:p=" << pr << pind(pr) << "\n";
  put_tpci(fout, opt, tp, vt, mcbon, opt.m(), ftab);
}

void wtestclass::put_tpci(ostream& fout, option& opt, double tp, double vt, mctype mc, int dc, const ftable& ftab) const
// Wilcoxonの1標本検定のTpの信頼区間出力
{
  fout << "  符号付き順位和の" << opt.cc() * 100.0 << "%";
  if (mc != mcfis) fout << "同時";
  fout << "信頼区間\n";
  double nf = ftab.ftotal();
  double tn = nf * (nf + 1.0) * 0.5;
  double tm = tn - tp;
  double ts = tp - tm;
  double ci = calmcci(mc, opt.cc(), 1, 1, 0, dc, 4.0 * vt);
  double tsl = ts - ci, tsu = ts + ci;
  if (tsl < -tn) tsl = -tn;
  if (tsu > tn) tsu = tn;
  double rs = ts / nf, cin = ci  / nf;
  double rsl = rs - cin, rsu = rs + cin;
  if (rsl < -tn / nf) rsl = -tn / nf;
  if (rsu > tn / nf) rsu = tn / nf;
  double pro = 100.0 / nf;
  fout << "  " << tp << "-" << tm << "=" << ts << "±" << ci << "(" << tsl << "-" << tsu << ")\n";
  fout << "  符号付き順位平均値=" << rs << "±" << cin << "(" << rsl << "-" << rsu << ")\n";
  fout << "  合計に対する割合(%)=" << rs * pro << "±" << cin * pro << "(" << rsl * pro << "-" << rsu * pro << ")\n";
}

void wtestclass::put_stest(ostream& fout, option& opt, const ftable& ftab) const
// 符号検定の結果を出力する
{
  ldbl pr1 = binom(ftab.frtotal(1), ftab.frtotal(2));
  ldbl pr = (pr1 < 0.5L) ? 2.0L * pr1 : 1.0L;
  fout << "\n・符号検定(理論確率=0.5の二項検定)\n";
  fout << "  有意確率:片側p=" << pr1 << pind(pr1) << " 両側p=" << pr << pind(pr) << "\n";
  put_stpci(fout, opt, mcfis, 1, ftab);
  if (opt.res("m") == false || opt.m() == 1) return;
  pr1 = pr1 * (ldbl)(opt.m());
  if (pr1 > 1.0L) pr1 = 1.0L;
  pr = (pr1 < 0.5L) ? 2.0L * pr1 : 1.0L;
  fout << "\n  Bonferroni型多重検定  有意確率:片側p=" << pr1 << pind(pr1) << " 両側p=" << pr << pind(pr) << "\n";
  put_stpci(fout, opt, mcbon, opt.m(), ftab);
}

void wtestclass::put_stpci(ostream& fout, option& opt, mctype mc, int dc, const ftable& ftab) const
// 符号検定のpp信頼区間出力
{
  fout << "  [" << ftab.rcode(1) << "の出現確率 - 理論確率]の" << opt.cc() * 100.0 << "%";
  if (mc != mcfis) fout << "同時";
  fout << "信頼区間\n";
  int nt = ftab.ftotal();
  int n = ftab.frtotal(1);
  double p0 = (double)n / (double)nt;
  double dp = p0 - 0.5;
  double al = 1.0 - opt.cc();
  double cc = 1.0 - al / (double)dc;
  double plow = 0.0, pup = 1.0;
  double pci = calbexci(plow, pup, n, nt, cc);
  fout << "  " << p0 << "-0.5=" << dp << "  下限=" << plow - 0.5 << "  上限=" << pup - 0.5 << endl;
}

void wtestclass::put_mtest(ostream& fout, option& opt, const ftable& ftab) const
// McNemar検定の結果を出力する
{
  double tp = (double)(ftab.frtotal(1) - ftab.frtotal(2));
  double vt = (double)(ftab.ftotal());
  double z = (tp - sgn(tp)) / sqrt(vt);
  if (opt.mt() == testorg) z = tp / sqrt(vt);
  ldbl pr = ztop(z);
  fout << "\n・McNemarの検定";
  if (opt.mt() != testorg) fout << "(修正有)";
  fout << "\n  正規分布:z=" << z << " 両側有意確率:p=" << pr << pind(pr) << "\n";
  put_mtpci(fout, opt, mcfis, 1, ftab);
  if (opt.res("m") == false || opt.m() == 1) return;
  pr = pr * (ldbl)(opt.m());
  if (pr > 1.0L) pr = 1.0L;
  fout << "\n  Bonferroni型多重検定  両側有意確率:p=" << pr << pind(pr) << "\n";
  put_mtpci(fout, opt, mcbon, opt.m(), ftab);
}

void wtestclass::put_mtpci(ostream& fout, option& opt, mctype mc, int dc, const ftable& ftab) const
// McNemarの検定のpp信頼区間出力
{
  fout << "  [" << ftab.rcode(1) << "の出現確率 - 理論確率]の" << opt.cc() * 100.0 << "%";
  if (mc != mcfis) fout << "同時";
  fout << "信頼区間\n";
  int nt = ftab.ftotal();
  double p0 = (double)(ftab.frtotal(1)) / (double)nt;
  double dp = p0 - 0.5;
  double al = 1.0 - opt.cc();
  double cc = 1.0 - al / (double)dc;
  bool cor = (opt.mt() == testorg) ? false : true;
  double plow = 0.0, pup = 1.0;
  double pci = calbnmci(plow, pup, p0, nt, cc, cor);
  fout << "  " << p0 << "-0.5=" << dp << "±" << pci << "(" << dp - pci << "-" << dp + pci << ")\n";
}

int wtestclass::get(const char* c_fname, option& opt) // 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("実行時エラー！");
  if (opt.check(nftab) != 0) throw option_error("実行時オプションエラー！");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void wtestclass::put(option& opt) // 結果を出力する
{
  if (opt.res("res") == false) return;
  listidd_itr iitr = idlst.begin();
  for (auto& ftabi : ftablst)
  {
    put_id(cout, *iitr);
    if (ftabi.nr() > 0)
    {
      if (opt.res("tab"))
      {
        if (opt.tab() == tabetc)
        {
          if (ftabi.nc() == 1) putftable(cout, ftabi, tabfprc);
          else putftable(cout, ftabi, tabfprr);
        }
        else
        { putftable(cout, ftabi, opt.tab()); }
      }
      if (ftabi.nr() == 2)
      {
        if (opt.res("wt")) put_wtest(cout, opt, ftabi);
        if (opt.res("st") || (opt.res("def") && ftabi.nc() == 1)) put_stest(cout, opt, ftabi);
        if (opt.res("mt")) put_mtest(cout, opt, ftabi);
        cout << "\n" << endl;
      }
    }
    ++iitr;
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  wtestclass wtest;             // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          wtest.get(nullptr, opt);
          wtest.put(opt);
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        wtest.get(argv[i], opt);  fok = false;
        wtest.put(opt);
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "内容不明なエラー発生！");
    return 255;
  }
  return 0;
}
